from http import HTTPStatus
from unittest.mock import AsyncMock, patch

from synapse.server import HomeServer
from synapse.util.clock import Clock
from twisted.internet.testing import MemoryReactor

from famedly_control_synapse.client import (
    DiffRecord,
    ManyGroupsDiffResponse,
    Membership,
)
from famedly_control_synapse.repository import ManagedRoomRepository
from famedly_control_synapse.rest.types import CreateManagedRoomRequest
from famedly_control_synapse.sync import GroupMembershipSyncer
from tests.utils.module_api_testcase import ModuleApiTestCase


class TestGroupMembershipSync(ModuleApiTestCase):
    def default_config(self):
        conf = super().default_config()
        conf["modules"][0]["config"]["sync_enabled"] = True
        return conf

    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)
        self.syncer: GroupMembershipSyncer = self.hs.room_control.syncer  # type: ignore[attr-defined]
        self.repository: ManagedRoomRepository = self.hs.room_control.repository  # type: ignore[attr-defined]
        self.member_1 = self.register_user("sync_member_1", "password")
        self.member_2 = self.register_user("sync_member_2", "password")
        self.member_3 = self.register_user("sync_member_3", "password")

    def _create_managed_room_for_sync(
        self, name: str = "Sync Test Room", groups: list[str] | None = None
    ) -> str:
        groups = groups or ["sync_group"]
        with (
            patch(
                "famedly_control_synapse.client.FamedlyControlClient.get_group_members",
                new_callable=AsyncMock,
                return_value=[],
            ),
            patch(
                "famedly_control_synapse.room_handler.ManagedRoomHandler.batch_convert_external_user_ids_to_matrix_user_ids",
                new_callable=AsyncMock,
                side_effect=lambda x: (x, []),
            ),
        ):
            config = CreateManagedRoomRequest(
                room_alias_name=f"sync_room_{self._room_counter + 1}",
                name=name,
                groups=groups,
            )
            self._room_counter += 1
            channel = self.make_request(
                method="POST",
                path=self.CREATE_PATH,
                content=config.model_dump(),
                access_token=self.creator_access_token,
                shorthand=False,
            )
            assert channel.code == HTTPStatus.OK, channel.result
            # The endpoint for room creation establishes the sync token for us. To allow
            # the process_sync() function to work, just fill it in the same way it is
            # done by start_sync_loop().
            entry = self.get_success(self.repository.get_sync_token_entry())
            assert entry is not None
            self.syncer._sync_token_user_id, self.syncer._sync_token = entry
            assert self.syncer._sync_token_user_id == self.creator
            return channel.json_body["room_id"]

    def _get_membership(self, room_id: str, user_id: str) -> str | None:
        path = f"/_matrix/client/v3/rooms/{room_id}/state/m.room.member/{user_id}"
        channel = self.make_request("GET", path, access_token=self.creator_access_token)
        if channel.code == HTTPStatus.OK:
            return channel.json_body["membership"]
        return None

    @patch(
        "famedly_control_synapse.room_handler.ManagedRoomHandler.batch_convert_external_user_ids_to_matrix_user_ids",
        new_callable=AsyncMock,
    )
    @patch(
        "famedly_control_synapse.client.FamedlyControlClient.get_all_groups_diffs",
        new_callable=AsyncMock,
    )
    def test_sync_adds_users_to_room(self, mock_get_diffs, mock_batch_convert) -> None:
        """Sync should add new users while leaving existing members unchanged."""
        room_id = self._create_managed_room_for_sync(groups=["group_a"])
        mock_batch_convert.side_effect = lambda x: (x, [])

        # First sync: add member_1 as an existing member
        mock_get_diffs.return_value = ManyGroupsDiffResponse(
            next_sync="1",
            data={
                "group_a": [
                    DiffRecord(user_id=self.member_1, action=Membership.ADD),
                ],
            },
        )
        self.get_success(self.syncer._process_sync())
        assert self._get_membership(room_id, self.member_1) == "join"

        # Second sync: add member_2 and member_3
        mock_get_diffs.return_value = ManyGroupsDiffResponse(
            next_sync="2",
            data={
                "group_a": [
                    DiffRecord(user_id=self.member_2, action=Membership.ADD),
                    DiffRecord(user_id=self.member_3, action=Membership.ADD),
                ],
            },
        )
        self.get_success(self.syncer._process_sync())

        # New users joined
        assert self._get_membership(room_id, self.member_2) == "join"
        assert self._get_membership(room_id, self.member_3) == "join"
        # Existing member unchanged
        assert self._get_membership(room_id, self.member_1) == "join"

    @patch(
        "famedly_control_synapse.room_handler.ManagedRoomHandler.batch_convert_external_user_ids_to_matrix_user_ids",
        new_callable=AsyncMock,
    )
    @patch(
        "famedly_control_synapse.client.FamedlyControlClient.get_all_groups_diffs",
        new_callable=AsyncMock,
    )
    def test_sync_removes_users_from_room(
        self, mock_get_diffs, mock_batch_convert
    ) -> None:
        """Sync should remove specified users while leaving others unchanged."""
        room_id = self._create_managed_room_for_sync(groups=["group_b"])
        mock_batch_convert.side_effect = lambda x: (x, [])

        # First add all three members
        mock_get_diffs.return_value = ManyGroupsDiffResponse(
            next_sync="1",
            data={
                "group_b": [
                    DiffRecord(user_id=self.member_1, action=Membership.ADD),
                    DiffRecord(user_id=self.member_2, action=Membership.ADD),
                    DiffRecord(user_id=self.member_3, action=Membership.ADD),
                ],
            },
        )
        self.get_success(self.syncer._process_sync())
        assert self._get_membership(room_id, self.member_1) == "join"
        assert self._get_membership(room_id, self.member_2) == "join"
        assert self._get_membership(room_id, self.member_3) == "join"

        # Now remove only member_1
        mock_get_diffs.return_value = ManyGroupsDiffResponse(
            next_sync="2",
            data={
                "group_b": [
                    DiffRecord(user_id=self.member_1, action=Membership.REM),
                ],
            },
        )
        self.get_success(self.syncer._process_sync())

        # Removed user is gone
        assert self._get_membership(room_id, self.member_1) == "leave"
        # Remaining members unchanged
        assert self._get_membership(room_id, self.member_2) == "join"
        assert self._get_membership(room_id, self.member_3) == "join"

    @patch(
        "famedly_control_synapse.room_handler.ManagedRoomHandler.batch_convert_external_user_ids_to_matrix_user_ids",
        new_callable=AsyncMock,
    )
    @patch(
        "famedly_control_synapse.client.FamedlyControlClient.get_all_groups_diffs",
        new_callable=AsyncMock,
    )
    def test_sync_mixed_add_and_remove(
        self, mock_get_diffs, mock_batch_convert
    ) -> None:
        """A single diff with both Add and Rem actions should add new users,
        remove specified users, and leave others unchanged."""
        room_id = self._create_managed_room_for_sync(groups=["group_mixed"])
        mock_batch_convert.side_effect = lambda x: (x, [])

        # Set up initial state: member_1 and member_2 are in the room
        mock_get_diffs.return_value = ManyGroupsDiffResponse(
            next_sync="1",
            data={
                "group_mixed": [
                    DiffRecord(user_id=self.member_1, action=Membership.ADD),
                    DiffRecord(user_id=self.member_2, action=Membership.ADD),
                ],
            },
        )
        self.get_success(self.syncer._process_sync())
        assert self._get_membership(room_id, self.member_1) == "join"
        assert self._get_membership(room_id, self.member_2) == "join"

        # Now add member_3 and remove member_1 in one diff
        mock_get_diffs.return_value = ManyGroupsDiffResponse(
            next_sync="2",
            data={
                "group_mixed": [
                    DiffRecord(user_id=self.member_3, action=Membership.ADD),
                    DiffRecord(user_id=self.member_1, action=Membership.REM),
                ],
            },
        )
        self.get_success(self.syncer._process_sync())

        # Newly added
        assert self._get_membership(room_id, self.member_3) == "join"
        # Removed
        assert self._get_membership(room_id, self.member_1) == "leave"
        # Unchanged
        assert self._get_membership(room_id, self.member_2) == "join"

    @patch(
        "famedly_control_synapse.room_handler.ManagedRoomHandler.batch_convert_external_user_ids_to_matrix_user_ids",
        new_callable=AsyncMock,
    )
    @patch(
        "famedly_control_synapse.client.FamedlyControlClient.get_all_groups_diffs",
        new_callable=AsyncMock,
    )
    def test_sync_advances_token_on_success(
        self, mock_get_diffs, mock_batch_convert
    ) -> None:
        """Sync token should advance after successful processing."""
        self._create_managed_room_for_sync(groups=["group_c"])
        mock_batch_convert.side_effect = lambda x: (x, [])

        self.syncer._sync_token = None
        mock_get_diffs.return_value = ManyGroupsDiffResponse(
            next_sync="42",
            data={},
        )
        self.get_success(self.syncer._process_sync())

        assert self.syncer._sync_token == "42"
        entry = self.get_success(self.repository.get_sync_token_entry())
        assert entry is not None
        assert entry[1] == "42"

    @patch(
        "famedly_control_synapse.client.FamedlyControlClient.get_all_groups_diffs",
        new_callable=AsyncMock,
    )
    def test_sync_token_persisted_and_restored(self, mock_get_diffs) -> None:
        """Sync token should be persisted to the database and restored on loop start."""
        # Initialize sync token account data by creating a managed room
        self._create_managed_room_for_sync(groups=["group_persist"])
        self.syncer._sync_token_user_id = self.creator
        self.syncer._sync_token = None
        mock_get_diffs.return_value = ManyGroupsDiffResponse(
            next_sync="99",
            data={},
        )
        self.get_success(self.syncer._process_sync())
        assert self.syncer._sync_token == "99"

        # Simulate restart: reset in-memory state and let start_sync_loop restore from DB
        self.syncer._sync_token = None
        self.syncer._sync_token_user_id = None
        with patch.object(self.syncer.api, "looping_background_call"):
            self.get_success(self.syncer.start_sync_loop())
        assert self.syncer._sync_token_user_id == self.creator
        assert self.syncer._sync_token == "99"

    @patch(
        "famedly_control_synapse.client.FamedlyControlClient.get_all_groups_diffs",
        new_callable=AsyncMock,
    )
    def test_sync_does_not_advance_token_on_failure(self, mock_get_diffs) -> None:
        """Sync token should NOT advance if membership updates fail."""
        self._create_managed_room_for_sync(groups=["group_d"])

        self.syncer._sync_token = "5"
        mock_get_diffs.return_value = ManyGroupsDiffResponse(
            next_sync="10",
            data={
                "group_d": [
                    DiffRecord(user_id=self.member_1, action=Membership.ADD),
                ],
            },
        )

        with (
            patch(
                "famedly_control_synapse.room_handler.ManagedRoomHandler.batch_convert_external_user_ids_to_matrix_user_ids",
                new_callable=AsyncMock,
                return_value=([self.member_1], []),
            ),
            patch(
                "famedly_control_synapse.room_handler.ManagedRoomHandler.force_join_users_to_room",
                new_callable=AsyncMock,
                return_value={self.member_1: "user not found"},
            ),
        ):
            self.get_success(self.syncer._process_sync())

        assert self.syncer._sync_token == "5"

    @patch(
        "famedly_control_synapse.client.FamedlyControlClient.get_all_groups_diffs",
        new_callable=AsyncMock,
    )
    def test_sync_retries_on_api_error(self, mock_get_diffs) -> None:
        """An API error should be caught by the sync loop without crashing."""
        mock_get_diffs.side_effect = Exception("connection refused")

        self.syncer._sync_token = None

        # _process_sync should propagate the exception (the loop catches it)
        self.get_failure(self.syncer._process_sync(), Exception)

        # Token should remain unchanged
        assert self.syncer._sync_token is None

    @patch(
        "famedly_control_synapse.room_handler.ManagedRoomHandler.batch_convert_external_user_ids_to_matrix_user_ids",
        new_callable=AsyncMock,
    )
    @patch(
        "famedly_control_synapse.client.FamedlyControlClient.get_all_groups_diffs",
        new_callable=AsyncMock,
    )
    def test_sync_handles_unknown_group(
        self, mock_get_diffs, mock_batch_convert
    ) -> None:
        """Diffs for groups not assigned to any managed room should be ignored."""
        self._create_managed_room_for_sync(groups=["group_e"])
        mock_batch_convert.side_effect = lambda x: (x, [])

        self.syncer._sync_token = None
        mock_get_diffs.return_value = ManyGroupsDiffResponse(
            next_sync="1",
            data={
                "unknown_group": [
                    DiffRecord(user_id=self.member_1, action=Membership.ADD),
                ],
            },
        )
        self.get_success(self.syncer._process_sync())

        # Token should still advance - unknown groups aren't errors
        assert self.syncer._sync_token == "1"

    @patch(
        "famedly_control_synapse.room_handler.ManagedRoomHandler.batch_convert_external_user_ids_to_matrix_user_ids",
        new_callable=AsyncMock,
    )
    @patch(
        "famedly_control_synapse.client.FamedlyControlClient.get_all_groups_diffs",
        new_callable=AsyncMock,
    )
    def test_sync_applies_diffs_to_multiple_rooms(
        self, mock_get_diffs, mock_batch_convert
    ) -> None:
        """If multiple rooms share the same group, diffs should apply to all of them."""
        room_id_1 = self._create_managed_room_for_sync(
            name="Room 1", groups=["shared_group"]
        )
        room_id_2 = self._create_managed_room_for_sync(
            name="Room 2", groups=["shared_group"]
        )
        mock_batch_convert.side_effect = lambda x: (x, [])

        mock_get_diffs.return_value = ManyGroupsDiffResponse(
            next_sync="1",
            data={
                "shared_group": [
                    DiffRecord(user_id=self.member_1, action=Membership.ADD),
                ],
            },
        )
        self.get_success(self.syncer._process_sync())

        assert self._get_membership(room_id_1, self.member_1) == "join"
        assert self._get_membership(room_id_2, self.member_1) == "join"

    @patch(
        "famedly_control_synapse.client.FamedlyControlClient.get_all_groups_diffs",
        new_callable=AsyncMock,
    )
    def test_sync_empty_data(self, mock_get_diffs) -> None:
        """Empty diff data should be a no-op and advance the token."""
        self.syncer._sync_token = None
        mock_get_diffs.return_value = ManyGroupsDiffResponse(
            next_sync="1",
            data={},
        )
        self.get_success(self.syncer._process_sync())
        assert self.syncer._sync_token == "1"



from http import HTTPStatus
from unittest.mock import AsyncMock, patch

from synapse.api.constants import CREATOR_POWER_LEVEL, EventTypes
from synapse.server import HomeServer
from synapse.util.clock import Clock
from twisted.internet.testing import MemoryReactor

from famedly_control_synapse.rest.types import CreateManagedRoomRequest
from tests.utils.module_api_testcase import ModuleApiTestCase


class TestPowerLevelHook(ModuleApiTestCase):
    CREATE_PATH = "/_famedlyControl/v1/managedRooms/createRoom"

    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)
        self.non_admin = self.register_user("non_admin", "password", admin=False)
        self.non_admin_token = self.login("non_admin", "password")

    def _create_managed_room(
        self, name: str = "PL Test Room", groups: list[str] | None = None
    ) -> str:
        config = CreateManagedRoomRequest(
            room_alias_name="pl_test_room",
            name=name,
            room_version="12",
            groups=groups or ["test_group"],
        )
        with (
            patch(
                "famedly_control_synapse.client.FamedlyControlClient.get_group_members",
                new_callable=AsyncMock,
                return_value=[],
            ),
            patch(
                "famedly_control_synapse.room_handler.ManagedRoomHandler.batch_convert_external_user_ids_to_matrix_user_ids",
                new_callable=AsyncMock,
                side_effect=lambda x: (x, []),
            ),
        ):
            channel = self.make_request(
                method="POST",
                path=self.CREATE_PATH,
                content=config.model_dump(),
                access_token=self.creator_access_token,
                shorthand=False,
            )
        assert channel.code == HTTPStatus.OK, channel.result
        return channel.json_body["room_id"]

    def _get_power_levels(self, room_id: str) -> dict:
        return self.helper.get_state(
            room_id, EventTypes.PowerLevels, self.creator_access_token
        )

    def _set_power_levels(
        self, room_id: str, content: dict, expect_code: int = HTTPStatus.OK
    ) -> dict:
        return self.helper.send_state(
            room_id,
            EventTypes.PowerLevels,
            content,
            self.creator_access_token,
            expect_code=expect_code,
        )

    def test_reject_other_user_equal_to_admin(self) -> None:
        """Setting another user's PL equal to the admin's should be rejected."""
        room_id = self._create_managed_room()
        pl = self._get_power_levels(room_id)

        pl["users"][self.non_admin] = CREATOR_POWER_LEVEL
        self._set_power_levels(room_id, pl, expect_code=HTTPStatus.FORBIDDEN)

    def test_reject_other_user_higher_than_admin(self) -> None:
        """Setting another user's PL higher than the admin's should be rejected."""
        room_id = self._create_managed_room()
        pl = self._get_power_levels(room_id)

        pl["users"][self.non_admin] = CREATOR_POWER_LEVEL + 1
        self._set_power_levels(room_id, pl, expect_code=HTTPStatus.FORBIDDEN)

    def test_reject_users_default_equal_to_admin(self) -> None:
        """Setting users_default equal to admin PL should be rejected."""
        room_id = self._create_managed_room()
        pl = self._get_power_levels(room_id)

        pl["users_default"] = CREATOR_POWER_LEVEL
        self._set_power_levels(room_id, pl, expect_code=HTTPStatus.FORBIDDEN)

    def test_reject_ban_level_below_non_admin(self) -> None:
        """Setting ban PL at or below a non-admin user's PL should be rejected."""
        room_id = self._create_managed_room()
        pl = self._get_power_levels(room_id)

        pl["users"][self.non_admin] = 50
        pl["ban"] = 50
        self._set_power_levels(room_id, pl, expect_code=HTTPStatus.FORBIDDEN)

    def test_reject_kick_level_below_non_admin(self) -> None:
        """Setting kick PL at or below a non-admin user's PL should be rejected."""
        room_id = self._create_managed_room()
        pl = self._get_power_levels(room_id)

        pl["users"][self.non_admin] = 50
        pl["kick"] = 50
        self._set_power_levels(room_id, pl, expect_code=HTTPStatus.FORBIDDEN)

    def test_reject_invite_level_below_non_admin(self) -> None:
        """Setting invite PL at or below a non-admin user's PL should be rejected."""
        room_id = self._create_managed_room()
        pl = self._get_power_levels(room_id)

        pl["users"][self.non_admin] = 50
        pl["invite"] = 50
        self._set_power_levels(room_id, pl, expect_code=HTTPStatus.FORBIDDEN)

    def test_reject_membership_level_below_users_default(self) -> None:
        """Membership action PLs must be above users_default too."""
        room_id = self._create_managed_room()
        pl = self._get_power_levels(room_id)

        pl["users_default"] = 30
        pl["ban"] = 30
        self._set_power_levels(room_id, pl, expect_code=HTTPStatus.FORBIDDEN)

    def test_allow_valid_power_level_change(self) -> None:
        """A valid PL change that maintains all constraints should be allowed."""
        room_id = self._create_managed_room()
        pl = self._get_power_levels(room_id)

        pl["users"][self.non_admin] = 10
        pl["ban"] = 100
        pl["kick"] = 100
        pl["invite"] = 100
        self._set_power_levels(room_id, pl, expect_code=HTTPStatus.OK)

    def test_allow_membership_levels_above_non_admin(self) -> None:
        """Membership action PLs strictly above all non-admin users should be allowed."""
        room_id = self._create_managed_room()
        pl = self._get_power_levels(room_id)

        pl["users"][self.non_admin] = 50
        pl["ban"] = 51
        pl["kick"] = 51
        pl["invite"] = 51
        self._set_power_levels(room_id, pl, expect_code=HTTPStatus.OK)

    def test_reject_admin_lower_than_other_user(self) -> None:
        """Admin lowering their PL below another user should be rejected."""
        room_id = self._create_managed_room()
        pl = self._get_power_levels(room_id)

        # First give non_admin a PL, then try to lower admin below it
        pl["users"][self.non_admin] = 50
        pl["ban"] = 51
        pl["kick"] = 51
        pl["invite"] = 51
        self._set_power_levels(room_id, pl, expect_code=HTTPStatus.OK)

        pl = self._get_power_levels(room_id)
        pl["users"][self.creator] = 49
        self._set_power_levels(room_id, pl, expect_code=HTTPStatus.FORBIDDEN)

    def test_reject_non_admin_at_sensitive_event_threshold(self) -> None:
        """Non-admin user with PL >= sensitive event threshold should be rejected,
        even if ban/kick/invite thresholds are set higher."""
        room_id = self._create_managed_room()
        pl = self._get_power_levels(room_id)

        # Set high ban/kick/invite thresholds to attempt bypass
        pl["ban"] = 1000
        pl["kick"] = 1000
        pl["invite"] = 1000
        # Give non-admin PL that is below membership thresholds but at state_default
        pl["users"][self.non_admin] = pl.get("state_default", 100)
        self._set_power_levels(room_id, pl, expect_code=HTTPStatus.FORBIDDEN)

    def test_admin_can_update_power_levels(self) -> None:
        """The room admin should be able to update power levels."""
        room_id = self._create_managed_room()
        pl = self._get_power_levels(room_id)

        pl["users"][self.non_admin] = 10
        self._set_power_levels(room_id, pl, expect_code=HTTPStatus.OK)

    def test_non_admin_cannot_update_power_levels(self) -> None:
        """A non-admin user in the room should not be able to update power levels."""
        room_id = self._create_managed_room()

        # Invite and join the non-admin to the room
        self.helper.invite(
            room_id,
            src=self.creator,
            targ=self.non_admin,
            tok=self.creator_access_token,
        )
        self.helper.join(room_id, user=self.non_admin, tok=self.non_admin_token)

        pl = self._get_power_levels(room_id)
        pl["users"][self.non_admin] = 10
        self.helper.send_state(
            room_id,
            EventTypes.PowerLevels,
            pl,
            self.non_admin_token,
            expect_code=HTTPStatus.FORBIDDEN,
        )

    def test_non_managed_room_not_affected(self) -> None:
        """Power level changes in non-managed rooms should not be affected by the hook."""
        room_id = self.helper.create_room_as(
            self.creator, tok=self.creator_access_token, is_public=False
        )
        pl = self.helper.get_state(
            room_id, EventTypes.PowerLevels, self.creator_access_token
        )

        pl["users"][self.non_admin] = pl["users"].get(self.creator, 100)
        self.helper.send_state(
            room_id,
            EventTypes.PowerLevels,
            pl,
            self.creator_access_token,
            expect_code=HTTPStatus.OK,
        )



from http import HTTPStatus
from unittest.mock import AsyncMock, patch

from synapse import event_auth
from synapse.api.constants import (
    CREATOR_POWER_LEVEL,
    EventTypes,
)
from synapse.api.errors import Codes, SynapseError
from synapse.server import HomeServer
from synapse.types.state import StateFilter
from synapse.util.clock import Clock
from twisted.internet.testing import MemoryReactor

from famedly_control_synapse.client import FamedlyControlError
from famedly_control_synapse.rest.room import MANAGED_ROOM_TYPE
from famedly_control_synapse.rest.types import CreateManagedRoomRequest, CreationContent
from famedly_control_synapse.room_handler import famedly_control_user_sync_error
from tests.utils.homeserver_testcase import override_config
from tests.utils.module_api_testcase import ModuleApiTestCase

CREATE_KEY = (EventTypes.Create, "")
POWER_KEY = (EventTypes.PowerLevels, "")


@patch(
    "famedly_control_synapse.client.FamedlyControlClient.get_group_members",
    new_callable=AsyncMock,
)
@patch(
    "famedly_control_synapse.room_handler.ManagedRoomHandler.batch_convert_external_user_ids_to_matrix_user_ids",
    new_callable=AsyncMock,
)
class TestManagedRoomCreation(ModuleApiTestCase):
    def room_config_v12(self):
        config = CreateManagedRoomRequest(
            room_alias_name="test_room_alias",
            name="Test Room",
            room_version="12",
            topic="This is a test room",
            groups=["test_group"],
        )
        return config.model_dump()

    def room_config_v10(self):
        config = CreateManagedRoomRequest(
            room_alias_name="test_room_alias",
            name="Test Room",
            room_version="10",
            creation_content=CreationContent(creator=self.creator),
            topic="This is a test room",
            groups=["test_group"],
        )
        return config.model_dump()

    def invalid_room_config(self):
        config = {
            "room_alias_name": "test_room_alias",
            "name": "Test Room",
            "topic": "This is a test room",
            "3pid_invites": ["something"],  # Invalid field
        }
        return config

    def test_room_creation_success(
        self, mock_batch_convert, mock_get_group_members
    ) -> None:
        """Tests that managed room creation returns the expected response"""
        mock_get_group_members.return_value = [
            self.invitee
        ]  # in real case this should be external_ids
        mock_batch_convert.side_effect = lambda x: (x, [])
        channel = self.make_request(
            method="POST",
            path=self.CREATE_PATH,
            content=self.room_config_v12(),
            access_token=self.creator_access_token,
            shorthand=False,
        )

        assert channel.code == HTTPStatus.OK, channel.result
        assert "room_id" in channel.json_body, "Response should contain room_id"
        room_id = channel.json_body["room_id"]

        # Check if the room has the correct configuration
        state_map = self.get_success(
            self.hs.get_storage_controllers().state.get_current_state(
                room_id,
                StateFilter.from_types(
                    [(EventTypes.GuestAccess, None), (EventTypes.JoinRules, None)]
                ),
            )
        )
        assert state_map[EventTypes.JoinRules, ""].content == {
            "join_rule": "invite"
        }, "Join rules should be set to invite"
        assert state_map[EventTypes.GuestAccess, ""].content == {
            "guest_access": "forbidden"
        }, "Guest access should be set to forbidden"

        # Check if account data is updated
        account_data = self.get_success(
            self.store.get_account_data_for_room(self.creator, room_id)
        )
        assert account_data == {MANAGED_ROOM_TYPE: {"groups": ["test_group"]}}

    def test_room_creation_invalid_body(
        self, mock_batch_convert, mock_get_group_members
    ) -> None:
        """Tests that invalid request body returns a 400 error with an error message"""
        channel = self.make_request(
            method="POST",
            path=self.CREATE_PATH,
            content=self.invalid_room_config(),
            access_token=self.creator_access_token,
            shorthand=False,
        )

        assert channel.code == HTTPStatus.BAD_REQUEST, channel.result
        assert (
            "Invalid request body" in channel.json_body["error"]
        ), "Response should contain an error message"

    def test_room_creation_with_missing_required_fields(
        self, mock_batch_convert, mock_get_group_members
    ) -> None:
        """Tests that missing required fields in the request body returns a 400 error with an error message"""
        channel = self.make_request(
            method="POST",
            path=self.CREATE_PATH,
            content={"name": "Test Room"},
            access_token=self.creator_access_token,
            shorthand=False,
        )

        assert channel.code == HTTPStatus.BAD_REQUEST, channel.result
        assert (
            "Invalid request body" in channel.json_body["error"]
        ), "Response should contain an error message"

    def test_room_creation_powerlevel_with_room_v12(
        self, mock_batch_convert, mock_get_group_members
    ) -> None:
        """Tests that the creator has the highest power level and no other user can have the same"""
        mock_get_group_members.return_value = [
            self.invitee
        ]  # in real case this should be external_ids
        mock_batch_convert.side_effect = lambda x: (x, [])
        room_config = self.room_config_v12()
        channel = self.make_request(
            method="POST",
            path=self.CREATE_PATH,
            content=room_config,
            access_token=self.creator_access_token,
            shorthand=False,
        )
        assert channel.code == HTTPStatus.OK, channel.result
        room_id = channel.json_body["room_id"]

        # Check if the creator has ultimate power level and no other user has the same power level
        auth_events = self.get_success(
            self.hs.get_storage_controllers().state.get_current_state(
                room_id,
                StateFilter.from_types(
                    [
                        POWER_KEY,
                        CREATE_KEY,
                    ]
                ),
            )
        )
        creator_pl = event_auth.get_user_power_level(self.creator, auth_events)
        assert (
            creator_pl == 9007199254740992
        ), "Creator should have power level 9007199254740992"

        # Check the invited user's power level
        invitee_pl = event_auth.get_user_power_level(self.invitee, auth_events)
        assert invitee_pl == 0, "Invitee should have power level 0"

    def test_room_creation_powerlevel_with_room_v10(
        self, mock_batch_convert, mock_get_group_members
    ) -> None:
        mock_get_group_members.return_value = [
            self.invitee
        ]  # in real case this should be external_ids
        mock_batch_convert.side_effect = lambda x: (x, [])
        room_config = self.room_config_v10()
        channel = self.make_request(
            method="POST",
            path=self.CREATE_PATH,
            content=room_config,
            access_token=self.creator_access_token,
            shorthand=False,
        )
        assert channel.code == HTTPStatus.OK, channel.result
        room_id = channel.json_body["room_id"]

        # Check if the creator has infinite power level and no other user has the same power level
        auth_events = self.get_success(
            self.hs.get_storage_controllers().state.get_current_state(
                room_id,
                StateFilter.from_types(
                    [
                        POWER_KEY,
                        CREATE_KEY,
                    ]
                ),
            )
        )
        creator_pl = event_auth.get_user_power_level(self.creator, auth_events)
        assert (
            creator_pl == CREATOR_POWER_LEVEL - 1
        ), "Creator should have infinite power level"

        # Check the invited user's power level
        invitee_pl = event_auth.get_user_power_level(self.invitee, auth_events)
        assert invitee_pl == 0, "Invitee should have power level 0"

    def test_room_created_with_members_joined(
        self, mock_batch_convert, mock_get_group_members
    ) -> None:
        """Tests that the users of the groups are joined to the room after creation"""
        test_group = "test_group_1"
        test_member_1 = self.register_user("test_member_1", "password")
        test_member_2 = self.register_user("test_member_2", "password")
        group_members = [test_member_1, test_member_2]
        mock_get_group_members.return_value = (
            group_members  # in real case this should be external_ids
        )
        mock_batch_convert.side_effect = lambda x: (x, [])

        # Create a managed room with a group that has members
        room_id = self._create_managed_room(
            name="Test Room with Group Members", groups=[test_group]
        )

        # Check if the member of the group is joined to the room
        for member in group_members:
            path = f"/_matrix/client/v3/rooms/{room_id}/state/m.room.member/{member}"
            channel = self.make_request(
                "GET", path, access_token=self.creator_access_token
            )
            assert (
                channel.code == HTTPStatus.OK
            ), f"Expected 200 but got {channel.code} for member {member}"
            assert channel.json_body["membership"] == "join", channel.json_body[
                "membership"
            ]

        # Check the account data is updated
        room_account_data = self.get_success(
            self.store.get_account_data_for_room(self.creator, room_id)
        )
        assert room_account_data == {
            MANAGED_ROOM_TYPE: {"groups": [test_group]}
        }, room_account_data

    @override_config({"run_background_tasks_on": "made_worker_name"})
    def test_non_background_worker_managed_room_creation_endpoint_disabled(
        self, mock_batch_convert, mock_get_group_members
    ) -> None:
        """
        Tests that the managed room creation endpoint does not function on the incorrect
        Synapse worker
        """
        mock_get_group_members.return_value = [
            self.invitee
        ]  # in real case this should be external_ids
        mock_batch_convert.side_effect = lambda x: (x, [])
        channel = self.make_request(
            method="POST",
            path=self.CREATE_PATH,
            content=self.room_config_v12(),
            access_token=self.creator_access_token,
            shorthand=False,
        )

        assert channel.code == HTTPStatus.NOT_FOUND, channel.result


@patch(
    "famedly_control_synapse.client.FamedlyControlClient.get_group_members",
    new_callable=AsyncMock,
)
@patch(
    "famedly_control_synapse.room_handler.ManagedRoomHandler.batch_convert_external_user_ids_to_matrix_user_ids",
    new_callable=AsyncMock,
)
class TestAssignGroupsToManagedRoom(ModuleApiTestCase):

    def test_update_single_group_to_single_group(
        self, mock_batch_convert, mock_get_group_members
    ) -> None:
        """Tests that managed room which already have groups can be updated and the
        account data is updated correctly."""
        test_old_group = "test_old_group"
        test_member_1 = self.register_user("test_member_1", "password")
        test_member_2 = self.register_user("test_member_2", "password")
        old_group_members = [test_member_1, test_member_2]

        test_new_group = "test_new_group"
        test_member_3 = self.register_user("test_member_3", "password")
        new_group_members = [test_member_2, test_member_3]

        def get_group_members(group_id):  # in real case this should return external_ids
            if group_id == test_old_group:
                return old_group_members
            elif group_id == test_new_group:
                return new_group_members
            return []

        mock_get_group_members.side_effect = get_group_members
        mock_batch_convert.side_effect = lambda x: (x, [])

        # Create a managed room with the old group information
        room_id = self._create_managed_room(
            name="Test Room with Group Members", groups=[test_old_group]
        )
        self._test_get_membership(room_id, old_group_members, expect_code=200)
        room_account_data = self.get_success(
            self.store.get_account_data_for_room(self.creator, room_id)
        )
        assert room_account_data == {
            MANAGED_ROOM_TYPE: {"groups": [test_old_group]}
        }, room_account_data

        # Now update the group information
        channel = self.make_request(
            method="POST",
            path=self.BASE_PATH + f"/{room_id}/groups",
            content={"groups": [test_new_group]},
            access_token=self.creator_access_token,
            shorthand=False,
        )
        assert channel.code == HTTPStatus.OK, channel.result

        # Check if the new member of the group is joined to the room
        for member in new_group_members:
            path = f"/_matrix/client/v3/rooms/{room_id}/state/m.room.member/{member}"
            channel = self.make_request(
                "GET", path, access_token=self.creator_access_token
            )
            assert (
                channel.code == HTTPStatus.OK
            ), f"Expected 200 but got {channel.code} for member {member}"
            assert (
                channel.json_body["membership"] == "join"
            ), f"Expected membership to be join but got {channel.json_body['membership']} for member {member}"

        # Check if the old member who is not in the new group is removed from the room
        path = f"/_matrix/client/v3/rooms/{room_id}/state/m.room.member/{test_member_1}"
        channel = self.make_request("GET", path, access_token=self.creator_access_token)
        assert (
            channel.code == HTTPStatus.OK
        ), f"Expected 200 but got {channel.code} for member {test_member_1}"
        assert (
            channel.json_body["membership"] == "leave"
        ), f"Expected membership to be leave but got {channel.json_body['membership']} for member {test_member_1}"

        # Check the account data is updated
        room_account_data = self.get_success(
            self.store.get_account_data_for_room(self.creator, room_id)
        )
        assert room_account_data == {
            MANAGED_ROOM_TYPE: {"groups": [test_new_group]}
        }, room_account_data

    def test_update_single_group_to_multiple_groups(
        self, mock_batch_convert, mock_get_group_members
    ) -> None:
        """
        Test updating a single group to multiple groups in a managed room.
        Old groups info: test_group_1
        New groups info: test_group_1, test_group_2, test_group_3
        Edge case to consider with member r who was in group 1 but removed from group later.
        """
        test_group_1 = "test_group_1"
        test_member_a = self.register_user("test_member_a", "password")
        test_member_b = self.register_user("test_member_b", "password")
        test_member_r = self.register_user("test_member_r", "password")
        test_group_1_current_members = [test_member_a, test_member_b]
        test_group_1_creation_members = [test_member_a, test_member_b, test_member_r]

        test_group_2 = "test_group_2"
        test_member_c = self.register_user("test_member_c", "password")
        test_group_2_members = [test_member_b, test_member_c]

        test_group_3 = "test_group_3"
        test_member_d = self.register_user("test_member_d", "password")
        test_member_e = self.register_user("test_member_e", "password")
        test_group_3_members = [test_member_d, test_member_e]

        old_group_info = [test_group_1]
        new_group_info = [test_group_1, test_group_2, test_group_3]

        call_counts = {}

        def get_members_by_group(
            group_id,
        ):  # in real case this should return external_ids
            if group_id not in call_counts:
                call_counts[group_id] = 0
            # Track how many times each group_id has been called
            call_counts[group_id] += 1

            if group_id == test_group_1:
                # First call returns creation members, subsequent calls return current members
                if call_counts[group_id] == 1:
                    return test_group_1_creation_members
                else:
                    return test_group_1_current_members
            elif group_id == test_group_2:
                return test_group_2_members
            elif group_id == test_group_3:
                return test_group_3_members
            return []

        mock_get_group_members.side_effect = get_members_by_group
        mock_batch_convert.side_effect = lambda x: (x, [])

        # Create a managed room with the old group information
        room_id = self._create_managed_room(
            name="Test Room with Group Members", groups=old_group_info
        )

        self._test_get_membership(
            room_id, test_group_1_creation_members, expect_code=200
        )
        room_account_data = self.get_success(
            self.store.get_account_data_for_room(self.creator, room_id)
        )
        assert room_account_data == {
            MANAGED_ROOM_TYPE: {"groups": old_group_info}
        }, room_account_data

        # Now update the group information
        channel = self.make_request(
            method="POST",
            path=self.BASE_PATH + f"/{room_id}/groups",
            content={"groups": new_group_info},
            access_token=self.creator_access_token,
            shorthand=False,
        )
        assert channel.code == HTTPStatus.OK, channel.result

        member_should_be_in_room = set(
            test_group_1_current_members + test_group_2_members + test_group_3_members
        )
        member_should_not_be_in_room = [test_member_r]

        # Check if the new member of the group is joined to the room
        for member in member_should_be_in_room:
            path = f"/_matrix/client/v3/rooms/{room_id}/state/m.room.member/{member}"
            channel = self.make_request(
                "GET", path, access_token=self.creator_access_token
            )
            assert (
                channel.code == HTTPStatus.OK
            ), f"Expected 200 but got {channel.code} for member {member}"
            assert (
                channel.json_body["membership"] == "join"
            ), f"Expected membership to be join but got {channel.json_body['membership']} for member {member}"

        for member in member_should_not_be_in_room:
            path = f"/_matrix/client/v3/rooms/{room_id}/state/m.room.member/{member}"
            channel = self.make_request(
                "GET", path, access_token=self.creator_access_token
            )
            assert (
                channel.code == HTTPStatus.OK
            ), f"Expected 200 but got {channel.code} for member {member}"
            assert (
                channel.json_body["membership"] == "leave"
            ), f"Expected membership to be leave but got {channel.json_body['membership']} for member {member}"

        # Check the account data is updated
        room_account_data = self.get_success(
            self.store.get_account_data_for_room(self.creator, room_id)
        )
        assert room_account_data == {
            MANAGED_ROOM_TYPE: {"groups": new_group_info}
        }, room_account_data

    def test_update_multiple_groups_to_multiple_groups(
        self, mock_batch_convert, mock_get_group_members
    ) -> None:
        """
        Test updating multiple groups to multiple groups in a managed room.
        Old groups info: test_group_1, test_group_2, test_group_3
        New groups info: test_group_1, test_group_4
        The edge cases to consider with member_r and member_b who are in multiple groups.
        """
        test_group_1 = "test_group_1"
        test_member_a = self.register_user("test_member_a", "password")
        test_member_b = self.register_user("test_member_b", "password")
        test_member_r = self.register_user("test_member_r", "password")
        test_group_1_current_members = [test_member_a, test_member_b]
        test_group_1_creation_members = [test_member_a, test_member_b, test_member_r]

        test_group_2 = "test_group_2"
        # Edge case example: member b is in both group 1 and 2. Group 1 is staying in
        # the room and group 2 is removed from the room.
        # member b should stay
        test_member_c = self.register_user("test_member_c", "password")
        test_group_2_members = [test_member_b, test_member_c]

        test_group_3 = "test_group_3"
        test_member_d = self.register_user("test_member_d", "password")
        test_member_e = self.register_user("test_member_e", "password")
        test_group_3_members = [test_member_d, test_member_e]

        test_group_4 = "test_group_4"
        # Edge case example: member r is removed from group 1, but added by group 4
        # and the group 1 remains to the room and group 4 is added to the room.
        # member r should stay
        test_member_f = self.register_user("test_member_f", "password")
        test_group_4_members = [test_member_f, test_member_r]

        old_group_info = [test_group_1, test_group_2, test_group_3]
        new_group_info = [test_group_1, test_group_4]

        call_counts = {}

        def get_members_by_group(
            group_id,
        ):  # in real case this should return external_ids
            if group_id not in call_counts:
                call_counts[group_id] = 0
            call_counts[group_id] += 1

            if group_id == test_group_1:
                # First call returns creation members, subsequent calls return current members
                if call_counts[group_id] == 1:
                    return test_group_1_creation_members
                else:
                    return test_group_1_current_members
            elif group_id == test_group_2:
                return test_group_2_members
            elif group_id == test_group_3:
                return test_group_3_members
            elif group_id == test_group_4:
                return test_group_4_members
            return []

        mock_get_group_members.side_effect = get_members_by_group
        mock_batch_convert.side_effect = lambda x: (x, [])

        # Create a managed room with the old group information
        room_id = self._create_managed_room(
            name="Test Room with Group Members", groups=old_group_info
        )
        self._test_get_membership(
            room_id,
            test_group_1_creation_members + test_group_2_members + test_group_3_members,
            expect_code=200,
        )
        room_account_data = self.get_success(
            self.store.get_account_data_for_room(self.creator, room_id)
        )
        assert room_account_data == {
            MANAGED_ROOM_TYPE: {"groups": old_group_info}
        }, room_account_data

        # Now update the group information
        channel = self.make_request(
            method="POST",
            path=self.BASE_PATH + f"/{room_id}/groups",
            content={"groups": new_group_info},
            access_token=self.creator_access_token,
            shorthand=False,
        )
        assert channel.code == HTTPStatus.OK, channel.result

        member_should_be_in_room = test_group_1_current_members + test_group_4_members
        member_should_not_be_in_room = [test_member_c] + test_group_3_members

        # Check if the new member of the group is joined to the room
        for member in member_should_be_in_room:
            path = f"/_matrix/client/v3/rooms/{room_id}/state/m.room.member/{member}"
            channel = self.make_request(
                "GET", path, access_token=self.creator_access_token
            )
            assert (
                channel.code == HTTPStatus.OK
            ), f"Expected 200 but got {channel.code} for member {member}"
            assert (
                channel.json_body["membership"] == "join"
            ), f"Expected membership to be join but got {channel.json_body['membership']} for member {member}"

        for member in member_should_not_be_in_room:
            path = f"/_matrix/client/v3/rooms/{room_id}/state/m.room.member/{member}"
            channel = self.make_request(
                "GET", path, access_token=self.creator_access_token
            )
            assert (
                channel.code == HTTPStatus.OK
            ), f"Expected 200 but got {channel.code} for member {member}"
            assert (
                channel.json_body["membership"] == "leave"
            ), f"Expected membership to be leave but got {channel.json_body['membership']} for member {member}"

        # Check the account data is updated
        room_account_data = self.get_success(
            self.store.get_account_data_for_room(self.creator, room_id)
        )
        assert room_account_data == {
            MANAGED_ROOM_TYPE: {"groups": new_group_info}
        }, room_account_data

    def test_prevent_room_creator_membership_change(
        self, mock_batch_convert, mock_get_group_members
    ) -> None:
        """Tests that the creator's membership is not changed even if the creator is in the group and the group is removed from the room."""
        group_including_creator = "test_group_including_creator"
        member_1 = self.register_user("test_member_1", "password")
        member_2 = self.register_user("test_member_2", "password")
        group_including_creator_members = [self.creator, member_1, member_2]

        new_group = "new_group"
        member_3 = self.register_user("test_member_3", "password")
        new_group_members = [member_3]

        def get_group_members(group_id):  # in real case this should return external_ids
            if group_id == group_including_creator:
                return group_including_creator_members
            elif group_id == new_group:
                return new_group_members
            return []

        mock_get_group_members.side_effect = get_group_members
        mock_batch_convert.side_effect = lambda x: (x, [])

        # Create a managed room with a group that has the creator as a member
        room_id = self._create_managed_room(
            name="Test Room with Group Members", groups=[group_including_creator]
        )

        # Now update the group information to remove the group from the room
        channel = self.make_request(
            method="POST",
            path=self.BASE_PATH + f"/{room_id}/groups",
            content={"groups": [new_group]},
            access_token=self.creator_access_token,
            shorthand=False,
        )
        assert channel.code == HTTPStatus.OK, channel.result

        # Check if the creator is still joined to the room
        path = f"/_matrix/client/v3/rooms/{room_id}/state/m.room.member/{self.creator}"
        channel = self.make_request("GET", path, access_token=self.creator_access_token)
        assert (
            channel.code == HTTPStatus.OK
        ), f"Expected 200 but got {channel.code} for member {self.creator}"
        assert (
            channel.json_body["membership"] == "join"
        ), f"Expected membership to be join but got {channel.json_body['membership']} for member {self.creator}"

    def test_client_error_response(
        self, mock_batch_convert, mock_get_group_members
    ) -> None:
        """Test that if client raises an error, the endpoint returns an error with an
        error message."""
        # Create a managed room
        test_group = "test_group"
        test_member = self.register_user("test_member", "password")
        mock_get_group_members.return_value = [test_member]
        mock_batch_convert.side_effect = lambda x: (x, [])
        room_id = self._create_managed_room(name="Test Room", groups=[test_group])

        # Try to update groups with the client raising an error
        mock_get_group_members.side_effect = FamedlyControlError(
            500, "Service unavailable"
        )
        channel = self.make_request(
            method="POST",
            path=self.BASE_PATH + f"/{room_id}/groups",
            content={"groups": ["new_group"]},
            access_token=self.creator_access_token,
            shorthand=False,
        )
        assert channel.code == 500, channel.result
        assert "Service unavailable" in channel.json_body["error"]

    def test_conversion_error_response(
        self, mock_batch_convert, mock_get_group_members
    ) -> None:
        """Test that if batch conversion raises an error, the endpoint returns an error
        with an error message."""
        # Create a managed room
        test_group = "test_group"
        test_member = self.register_user("test_member", "password")
        mock_get_group_members.return_value = [test_member]
        mock_batch_convert.side_effect = lambda x: (x, [])
        room_id = self._create_managed_room(name="Test Room", groups=[test_group])

        # Try to update groups and
        # batch_convert_external_user_ids_to_matrix_user_ids will return not_founds
        mock_batch_convert.side_effect = lambda x: (
            [],
            ["external_id_1", "external_id_2"],
        )
        channel = self.make_request(
            method="POST",
            path=self.BASE_PATH + f"/{room_id}/groups",
            content={"groups": ["new_group"]},
            access_token=self.creator_access_token,
            shorthand=False,
        )

        assert channel.code == HTTPStatus.BAD_REQUEST, channel.result
        assert (
            "Some external user IDs could not be mapped to Matrix user IDs"
            in channel.json_body["error"]
        )
        assert channel.json_body["details"] == ["external_id_1", "external_id_2"]

    def test_leave_error_all_fail_response(
        self,
        mock_batch_convert,
        mock_get_group_members,
    ) -> None:
        """Test that if leaving room raises errors for different users with different
        error codes, the endpoint returns the details and metrics are incremented for
        each error code.
        """
        # Get the metric initial values for both error codes
        server_name = self.hs.hostname
        initial_value_unknown = famedly_control_user_sync_error.labels(
            error_code="M_UNKNOWN", server_name=server_name
        )._value.get()
        initial_value_unauthorized = famedly_control_user_sync_error.labels(
            error_code="M_UNAUTHORIZED", server_name=server_name
        )._value.get()

        # Create a managed room with test_group_1
        test_group_1 = "test_group_1"
        test_member_1 = self.register_user("test_member_1", "password")
        test_member_2 = self.register_user("test_member_2", "password")
        mock_get_group_members.return_value = [test_member_1, test_member_2]
        mock_batch_convert.side_effect = lambda x: (x, [])
        room_id = self._create_managed_room(name="Test Room", groups=[test_group_1])

        # Prepare test_group_2 which removes all 2 members and add a new member
        test_group_2 = "test_group_2"
        test_member_3 = self.register_user("test_member_3", "password")
        mock_get_group_members.side_effect = lambda group_id: (
            [test_member_1, test_member_2]
            if group_id == test_group_1
            else [test_member_3]
        )

        # Try to update groups with update_room_membership raising errors for different users
        def update_room_membership_side_effect(
            sender, target, room_id, new_membership, **kwargs
        ):
            if target == test_member_1:
                raise SynapseError(500, "Unexpected error", Codes.UNKNOWN)
            elif target == test_member_2:
                raise SynapseError(401, "Permission denied", Codes.UNAUTHORIZED)
            return None

        with patch(
            "synapse.module_api.ModuleApi.update_room_membership",
            new_callable=AsyncMock,
        ) as mock_update:
            mock_update.side_effect = update_room_membership_side_effect

            channel = self.make_request(
                method="POST",
                path=self.BASE_PATH + f"/{room_id}/groups",
                content={"groups": [test_group_2]},
                access_token=self.creator_access_token,
                shorthand=False,
            )
            assert channel.code == 207, channel.result
            assert (
                "Failed to remove some members from the room"
                in channel.json_body["error"]
            )
            assert test_member_1 in channel.json_body["details"]
            assert test_member_2 in channel.json_body["details"]

        # Since both members failed to be removed, they should still be in the room
        # and the new member should be joined to the room
        for member in [test_member_1, test_member_2, test_member_3]:
            path = f"/_matrix/client/v3/rooms/{room_id}/state/m.room.member/{member}"
            channel = self.make_request(
                "GET", path, access_token=self.creator_access_token
            )
            assert (
                channel.code == HTTPStatus.OK
            ), f"Expected 200 but got {channel.code} for member {member}"
            assert (
                channel.json_body["membership"] == "join"
            ), f"Expected membership to be join but got {channel.json_body['membership']} for member {member}"

        # Check both metrics are incremented by 1 each
        new_value_unknown = famedly_control_user_sync_error.labels(
            error_code="M_UNKNOWN", server_name=server_name
        )._value.get()
        new_value_unauthorized = famedly_control_user_sync_error.labels(
            error_code="M_UNAUTHORIZED", server_name=server_name
        )._value.get()
        assert new_value_unknown == initial_value_unknown + 1
        assert new_value_unauthorized == initial_value_unauthorized + 1

    def test_join_error_response(
        self,
        mock_batch_convert,
        mock_get_group_members,
    ) -> None:
        """Test that if joining room raises an error, the endpoint returns an error with
        an error message and error metric was sent."""
        # Get the metric initial value
        server_name = self.hs.hostname
        initial_value = famedly_control_user_sync_error.labels(
            error_code="M_NOT_FOUND", server_name=server_name
        )._value.get()

        # Create a managed room with test_group_1 which has 2 valid members.
        test_group_1 = "test_group_1"
        test_group_2 = "test_group_2"
        test_member_1 = self.register_user("test_member_1", "password")
        test_member_2 = self.register_user("test_member_2", "password")
        non_existent_user = f"@random_user:{server_name}"
        mock_get_group_members.side_effect = lambda group_id: (
            [test_member_1, test_member_2]
            if group_id == test_group_1
            else [test_member_2, non_existent_user]
        )
        mock_batch_convert.side_effect = lambda x: (x, [])
        room_id = self._create_managed_room(name="Test Room", groups=[test_group_1])

        # Now update the group assignment to test_group_2, which has an already-joined
        # member and one non-existent member. `force_join_users_to_room` will skip for
        # the already-joined member and return error record for the non-existent member.
        def update_room_membership_side_effect(
            sender, target, room_id, new_membership, **kwargs
        ):
            if target == non_existent_user:
                raise SynapseError(404, "User not found", Codes.NOT_FOUND)
            return None

        with patch(
            "synapse.module_api.ModuleApi.update_room_membership",
            new_callable=AsyncMock,
        ) as mock_update:
            mock_update.side_effect = update_room_membership_side_effect

            channel = self.make_request(
                method="POST",
                path=self.BASE_PATH + f"/{room_id}/groups",
                content={"groups": [test_group_2]},
                access_token=self.creator_access_token,
                shorthand=False,
            )
            assert channel.code == 207, channel.result
            assert (
                "Failed to add some members to the room" in channel.json_body["error"]
            )
            assert non_existent_user in channel.json_body["details"]
            # test_member_2 should not be in details as the member was already in the room
            assert test_member_2 not in channel.json_body["details"]

        # Check the metric value is increased by 1
        new_value = famedly_control_user_sync_error.labels(
            error_code="M_NOT_FOUND", server_name=server_name
        )._value.get()
        assert new_value == initial_value + 1

    def test_both_join_and_leave_error_response(
        self,
        mock_batch_convert,
        mock_get_group_members,
    ) -> None:
        """Test that if both joining and leaving room raise errors, the endpoint returns
        an error message with both leave_errors and join_errors."""
        # Create a managed room with test_group_1
        test_group_1 = "test_group_1"
        test_group_2 = "test_group_2"
        test_member_1 = self.register_user("test_member_1", "password")
        test_member_2 = self.register_user("test_member_2", "password")
        non_existent_user = f"@random_user:{self.hs.hostname}"
        mock_get_group_members.side_effect = lambda group_id: (
            [test_member_1, test_member_2]
            if group_id == test_group_1
            else [non_existent_user]
        )
        mock_batch_convert.side_effect = lambda x: (x, [])
        room_id = self._create_managed_room(name="Test Room", groups=[test_group_1])

        # Now update the group assignment to test_group_2, which only has non-existent member.
        # Both `force_join_users_to_room` and `remove_users_from_room` will return errors.
        with (
            patch(
                "famedly_control_synapse.room_handler.ManagedRoomHandler.force_join_users_to_room",
                new_callable=AsyncMock,
            ) as mock_force_join,
            patch(
                "famedly_control_synapse.room_handler.ManagedRoomHandler.remove_users_from_room",
                new_callable=AsyncMock,
            ) as mock_remove,
        ):
            mock_force_join.return_value = {non_existent_user: "User not found"}
            mock_remove.return_value = {test_member_1: "Unexpected error"}

            channel = self.make_request(
                method="POST",
                path=self.BASE_PATH + f"/{room_id}/groups",
                content={"groups": [test_group_2]},
                access_token=self.creator_access_token,
                shorthand=False,
            )
            assert channel.code == 207, channel.result
            assert (
                "Failed to add and remove some members from the room"
                in channel.json_body["error"]
            )
            assert "leave_errors" in channel.json_body["details"]
            assert "join_errors" in channel.json_body["details"]
            assert test_member_1 in channel.json_body["details"]["leave_errors"]
            assert non_existent_user in channel.json_body["details"]["join_errors"]

    @override_config({"run_background_tasks_on": "fake_worker"})
    def test_non_background_worker_assign_groups_endpoint_disabled(
        self, mock_batch_convert, mock_get_group_members
    ) -> None:
        """
        Tests that the assign groups endpoint does not function on the incorrect
        Synapse worker
        """
        channel = self.make_request(
            method="POST",
            path=self.BASE_PATH + "/!fake_room_does_not_matter/groups",
            content={"groups": ["test_new_group"]},
            access_token=self.creator_access_token,
            shorthand=False,
        )
        assert channel.code == HTTPStatus.NOT_FOUND, channel.result


@patch(
    "famedly_control_synapse.client.FamedlyControlClient.get_group_members",
    new_callable=AsyncMock,
)
@patch(
    "famedly_control_synapse.room_handler.ManagedRoomHandler.batch_convert_external_user_ids_to_matrix_user_ids",
    new_callable=AsyncMock,
)
class TestListManagedRooms(ModuleApiTestCase):
    def prepare(self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer):
        super().prepare(reactor, clock, homeserver)
        self.non_admin = self.register_user("non_admin", "password", admin=False)
        self.non_admin_token = self.login("non_admin", "password")
        self.account_data_handler = homeserver.get_account_data_handler()

    def test_list_requires_admin(
        self, mock_batch_convert, mock_get_group_members
    ) -> None:
        """Non-admin users should get a 403."""
        channel = self.make_request(
            method="GET",
            path=self.LIST_PATH,
            access_token=self.non_admin_token,
            shorthand=False,
        )
        assert channel.code == HTTPStatus.FORBIDDEN, channel.result

    def test_list_requires_auth(
        self, mock_batch_convert, mock_get_group_members
    ) -> None:
        """Unauthenticated requests should get a 401."""
        channel = self.make_request(
            method="GET",
            path=self.LIST_PATH,
            shorthand=False,
        )
        assert channel.code == HTTPStatus.UNAUTHORIZED, channel.result

    def test_list_empty(self, mock_batch_convert, mock_get_group_members) -> None:
        """Listing when no managed rooms exist should return empty chunk."""
        channel = self.make_request(
            method="GET",
            path=self.LIST_PATH,
            access_token=self.creator_access_token,
            shorthand=False,
        )
        assert channel.code == HTTPStatus.OK, channel.result
        assert channel.json_body["chunk"] == []
        assert channel.json_body["total_room_count_estimate"] == 0

    def test_list_returns_managed_rooms(
        self, mock_batch_convert, mock_get_group_members
    ) -> None:
        """Created managed rooms should appear in the listing."""
        user_1 = self.register_user("user1", "password")
        user_2 = self.register_user("user2", "password")

        def get_members_by_group(
            group_id,
        ):  # in real case this should return external_ids
            if group_id == "group1":
                return [user_1]
            if group_id == "group2":
                return [user_2]
            return []

        mock_get_group_members.side_effect = get_members_by_group
        mock_batch_convert.side_effect = lambda x: (x, [])

        room_id = self._create_managed_room(
            name="Listed Room", groups=["group1", "group2"]
        )

        channel = self.make_request(
            method="GET",
            path=self.LIST_PATH,
            access_token=self.creator_access_token,
            shorthand=False,
        )
        assert channel.code == HTTPStatus.OK, channel.result
        assert channel.json_body["total_room_count_estimate"] == 1
        assert len(channel.json_body["chunk"]) == 1

        room_chunk = channel.json_body["chunk"][0]
        assert room_chunk["room_id"] == room_id
        assert room_chunk["de.famedly.managedRoom"]["groups"] == [
            "group1",
            "group2",
        ]

    def test_list_pagination(self, mock_batch_convert, mock_get_group_members) -> None:
        """Pagination should work with from and limit params."""
        user_1 = self.register_user("user1", "password")
        user_2 = self.register_user("user2", "password")
        user_3 = self.register_user("user3", "password")

        def get_members_by_group(
            group_id,
        ):  # in real case this should return external_ids
            if group_id == "group1":
                return [user_1]
            if group_id == "group2":
                return [user_2]
            if group_id == "group3":
                return [user_3]
            return []

        mock_get_group_members.side_effect = get_members_by_group
        mock_batch_convert.side_effect = lambda x: (x, [])

        room_ids = []
        for i in range(3):
            room_ids.append(
                self._create_managed_room(name=f"Room {i}", groups=[f"group{i}"])
            )

        # Get first page with limit 2
        channel = self.make_request(
            method="GET",
            path=f"{self.LIST_PATH}?limit=2",
            access_token=self.creator_access_token,
            shorthand=False,
        )
        assert channel.code == HTTPStatus.OK, channel.result
        assert len(channel.json_body["chunk"]) == 2
        assert channel.json_body["total_room_count_estimate"] == 3
        assert "next_batch" in channel.json_body
        assert "prev_batch" not in channel.json_body

        # Get second page
        next_batch = channel.json_body["next_batch"]
        channel = self.make_request(
            method="GET",
            path=f"{self.LIST_PATH}?from={next_batch}&limit=2",
            access_token=self.creator_access_token,
            shorthand=False,
        )
        assert channel.code == HTTPStatus.OK, channel.result
        assert len(channel.json_body["chunk"]) == 1
        assert "next_batch" not in channel.json_body
        assert "prev_batch" in channel.json_body

    @override_config({"run_background_tasks_on": "fake_worker"})
    def test_non_background_worker_list_endpoint_disabled(
        self, mock_batch_convert, mock_get_group_members
    ) -> None:
        """
        Tests that the assign groups endpoint does not function on the incorrect
        Synapse worker
        """
        channel = self.make_request(
            method="GET",
            path=self.LIST_PATH,
            access_token=self.creator_access_token,
            shorthand=False,
        )
        assert channel.code == HTTPStatus.NOT_FOUND, channel.result



# Copyright (C) 2026 Famedly
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import pytest
from pydantic import HttpUrl, ValidationError

from famedly_control_synapse.famedly_control import FamedlyControl


class TestConfigParsing:
    def test_valid_config(self):
        """Test parsing a valid config with all required fields."""
        config_dict = {
            "famedly_control": {
                "api_url": "https://api.example.com",
                "access_token": "test_token_123",
            },
            "auth_provider": "https://idp.example.com/",
        }
        config = FamedlyControl.parse_config(config_dict)

        assert config.famedly_control.api_url == HttpUrl("https://api.example.com")
        assert config.famedly_control.access_token == "test_token_123"
        assert config.auth_provider == "https://idp.example.com/"

    def test_missing_famedly_control(self):
        """Test that missing famedly_control raises ValidationError."""
        config_dict = {
            "auth_provider": "https://idp.example.com/",
        }

        with pytest.raises(ValidationError):
            FamedlyControl.parse_config(config_dict)

    def test_missing_api_url(self):
        """Test that missing api_url raises ValidationError."""
        config_dict = {
            "famedly_control": {
                "access_token": "test_token",
            },
            "auth_provider": "https://idp.example.com/",
        }

        with pytest.raises(ValidationError):
            FamedlyControl.parse_config(config_dict)

    def test_missing_access_token(self):
        """Test that missing access_token raises ValidationError."""
        config_dict = {
            "famedly_control": {
                "api_url": "https://api.example.com",
            },
            "auth_provider": "https://idp.example.com/",
        }

        with pytest.raises(ValidationError):
            FamedlyControl.parse_config(config_dict)

    def test_invalid_url_scheme(self):
        """Test that non-HTTP/HTTPS URL raises ValidationError."""
        config_dict = {
            "famedly_control": {
                "api_url": "ftp://example.com",
                "access_token": "test_token",
            },
            "auth_provider": "https://idp.example.com/",
        }

        with pytest.raises(ValidationError):
            FamedlyControl.parse_config(config_dict)

    def test_empty_access_token(self):
        """Test that empty access_token raises ValidationError."""
        config_dict = {
            "famedly_control": {
                "api_url": "https://api.example.com",
                "access_token": "",
            },
            "auth_provider": "https://idp.example.com/",
        }

        with pytest.raises(ValidationError):
            FamedlyControl.parse_config(config_dict)

    def test_whitespace_only_access_token(self):
        """Test that whitespace-only access_token raises ValidationError."""
        config_dict = {
            "famedly_control": {
                "api_url": "https://api.example.com",
                "access_token": "   ",
            },
            "auth_provider": "https://idp.example.com/",
        }

        with pytest.raises(ValidationError):
            FamedlyControl.parse_config(config_dict)



import json

from synapse.module_api import ModuleApi
from synapse.storage.engines import PostgresEngine
from synapse.types import JsonDict

from famedly_control_synapse.rest.types import MANAGED_ROOM_TYPE, SYNC_TOKEN_TYPE


class ManagedRoomRepository:
    def __init__(self, api: ModuleApi) -> None:
        self._api = api
        self._store = api._store
        self._account_data_handler = api._account_data_handler
        self._using_postgres = isinstance(self._store.db_pool.engine, PostgresEngine)

    async def count_managed_rooms(self) -> int:
        rows = await self._store.db_pool.execute(
            "count_managed_rooms",
            "SELECT COUNT(DISTINCT room_id) FROM room_account_data WHERE account_data_type = ?",
            MANAGED_ROOM_TYPE,
        )
        return rows[0][0] if rows else 0

    async def get_managed_rooms_paginated(
        self, limit: int, offset: int
    ) -> list[JsonDict]:
        if self._using_postgres:
            sql = """
                SELECT
                    paged.room_id,
                    paged.merged_groups,
                    rs.name,
                    rs.topic,
                    rs.canonical_alias,
                    rs.avatar,
                    rs.guest_access,
                    rs.history_visibility,
                    rc.joined_members
                FROM (
                    SELECT
                        ad.room_id,
                        JSONB_AGG(DISTINCT g.value) AS merged_groups

                    FROM room_account_data ad
                    LEFT JOIN jsonb_array_elements_text(ad.content::jsonb -> 'groups') g(value) ON TRUE
                    WHERE ad.account_data_type = ?
                    GROUP BY ad.room_id
                    ORDER BY ad.room_id
                    LIMIT ? OFFSET ?
                ) paged
                LEFT JOIN room_stats_state rs ON rs.room_id = paged.room_id
                LEFT JOIN room_stats_current rc ON rc.room_id = paged.room_id
            """
        else:
            sql = """
                SELECT
                    paged.room_id,
                    paged.merged_groups,
                    rs.name,
                    rs.topic,
                    rs.canonical_alias,
                    rs.avatar,
                    rs.guest_access,
                    rs.history_visibility,
                    rc.joined_members
                FROM (
                    SELECT
                        ad.room_id,
                        JSON_GROUP_ARRAY(DISTINCT g.value) AS merged_groups
                    FROM room_account_data ad
                    LEFT JOIN json_each(json_extract(ad.content, '$.groups')) g ON TRUE
                    WHERE ad.account_data_type = ?
                    GROUP BY ad.room_id
                    ORDER BY ad.room_id
                    LIMIT ? OFFSET ?
                ) paged
                LEFT JOIN room_stats_state rs ON rs.room_id = paged.room_id
                LEFT JOIN room_stats_current rc ON rc.room_id = paged.room_id
            """
        rows = await self._store.db_pool.execute(
            "get_managed_rooms_paginated",
            sql,
            MANAGED_ROOM_TYPE,
            limit,
            offset,
        )

        return [self._row_to_room_entry(row) for row in rows]

    async def get_rooms_by_group(self) -> dict[str, list[tuple[str, str]]]:
        """Get a mapping of group_id to list of (room_id, admin_user_id).

        The admin is the sender of the m.room.create event.
        Only rooms where the account_data user_id matches the room creator are returned.

        Returns:
            Dict mapping group IDs to lists of (room_id, admin_user_id) tuples.
        """
        rows = await self._store.db_pool.execute(
            "get_all_managed_rooms_with_groups",
            """SELECT ad.room_id, e.sender, ad.content
            FROM room_account_data ad
            INNER JOIN current_state_events cse
                ON cse.room_id = ad.room_id
                AND cse.type = 'm.room.create'
                AND cse.state_key = ''
            INNER JOIN events e
                ON e.event_id = cse.event_id
            WHERE ad.account_data_type = ?
                AND ad.user_id = e.sender""",
            MANAGED_ROOM_TYPE,
        )

        result: dict[str, list[tuple[str, str]]] = {}
        for room_id, sender, content_raw in rows:
            content = (
                json.loads(content_raw) if isinstance(content_raw, str) else content_raw
            )
            groups = content.get("groups", [])
            for group_id in groups:
                result.setdefault(group_id, []).append((room_id, sender))

        return result

    async def get_sync_token_entry(self) -> tuple[str, str | None] | None:
        """Load the persisted sync token and its owning user.

        If multiple entries exist, take any owned by a server admin.

        Returns:
            A (user_id, token) tuple, or None if no entry exists.
            The token itself may be None if this is the initial state.
        """
        rows = await self._store.db_pool.execute(
            "get_sync_token_entry",
            "SELECT user_id, content FROM account_data WHERE account_data_type = ?",
            SYNC_TOKEN_TYPE,
        )
        if not rows:
            return None

        for user_id, content_raw in rows:
            if await self._api.is_user_admin(user_id):
                content = (
                    json.loads(content_raw)
                    if isinstance(content_raw, str)
                    else content_raw
                )
                return user_id, content.get("token")

        # no admin found associated with SYNC_TOKEN_TYPE
        return None

    async def set_sync_token(self, user_id: str, token: str | None) -> None:
        """Persist the sync token as global account data for the given user."""
        await self._account_data_handler.add_account_data_for_user(
            user_id, SYNC_TOKEN_TYPE, {"token": token}
        )

    async def initialize_sync_token(self, user_id: str) -> None:
        """Create the sync token account data if it doesn't exist yet."""
        existing = await self.get_sync_token_entry()
        if existing is None:
            await self.set_sync_token(user_id, None)

    # TODO: add caching for this since it's called for every membership event
    async def is_managed_room(self, room_id: str, admin_user_id: str) -> bool:
        rows = await self._store.db_pool.execute(
            "is_managed_room",
            """SELECT 1 FROM room_account_data WHERE
            user_id = ? AND
            room_id = ? AND
            account_data_type = ?
            LIMIT 1""",
            admin_user_id,
            room_id,
            MANAGED_ROOM_TYPE,
        )
        return bool(rows)

    @staticmethod
    def _row_to_room_entry(row: tuple) -> JsonDict:
        (
            room_id,
            groups_raw,
            name,
            topic,
            canonical_alias,
            avatar,
            guest_access,
            history_visibility,
            joined_members,
        ) = row

        if groups_raw:
            groups = (
                json.loads(groups_raw) if isinstance(groups_raw, str) else groups_raw
            )
            groups = sorted(groups)
        else:
            groups = []

        entry: JsonDict = {
            "room_id": room_id,
            MANAGED_ROOM_TYPE: {"groups": groups},
        }
        if name:
            entry["name"] = name
        if topic:
            entry["topic"] = topic
        if canonical_alias:
            entry["canonical_alias"] = canonical_alias
        if avatar:
            entry["avatar_url"] = avatar
        if history_visibility is not None:
            entry["world_readable"] = history_visibility == "world_readable"
        if guest_access is not None:
            entry["guest_can_join"] = guest_access == "can_join"
        if joined_members is not None:
            entry["num_joined_members"] = joined_members

        return entry



# Copyright (C) 2026 Famedly
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
from pydantic import BaseModel, Field, HttpUrl, ValidationInfo, field_validator


def _validate_not_blank(v: str, info: ValidationInfo) -> str:
    if not v.strip():
        raise ValueError(f"{info.field_name} cannot be empty or whitespace-only")
    return v


class FamedlyControlApiConfig(BaseModel):
    api_url: HttpUrl = Field(
        ..., description="HTTP or HTTPS URL for the Famedly Control API"
    )
    access_token: str = Field(
        ...,
        min_length=1,
        description="Access token to authenticate against Famedly Control",
    )

    @field_validator("access_token")
    @classmethod
    def validate_token_fields(cls, v: str, info: ValidationInfo) -> str:
        return _validate_not_blank(v, info)


class FamedlyControlConfig(BaseModel):
    famedly_control: FamedlyControlApiConfig
    sync_enabled: bool = Field(
        default=True,
        description="Whether to run the background group membership sync loop",
    )
    sync_polling_interval_seconds: int = Field(
        default=30,
        ge=1,
        description="Delay in seconds before retrying after a sync failure",
    )
    auth_provider: str = Field(
        ...,
        description="The unique, internal ID of the external identity provider, used in the database to link external user IDs to Matrix user IDs",
    )

    @field_validator("auth_provider")
    @classmethod
    def validate_auth_provider(cls, v: str, info: ValidationInfo) -> str:
        return _validate_not_blank(v, info)



# Copyright (C) 2026 Famedly
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import logging

from synapse.module_api import ModuleApi

from famedly_control_synapse.client import DiffRecord, FamedlyControlClient, Membership
from famedly_control_synapse.config import FamedlyControlConfig
from famedly_control_synapse.repository import ManagedRoomRepository
from famedly_control_synapse.room_handler import ManagedRoomHandler

logger = logging.getLogger(__name__)


class GroupMembershipSyncer:
    """Background task that polls Famedly Control for group membership changes
    and updates managed room memberships accordingly."""

    def __init__(
        self,
        api: ModuleApi,
        client: FamedlyControlClient,
        room_handler: ManagedRoomHandler,
        repository: ManagedRoomRepository,
        config: FamedlyControlConfig,
    ) -> None:
        self.api = api
        self.client = client
        self.room_handler = room_handler
        self.repository = repository
        self.server_name = api.server_name
        self.polling_interval_seconds = config.sync_polling_interval_seconds
        self.is_enabled = config.sync_enabled
        self._sync_token: str | None = None
        self._sync_token_user_id: str | None = None
        self._is_running = False

    def start(self) -> None:
        """Start the periodic background sync.

        Safe to call multiple times, does nothing if already running.
        Schedules ``start_sync_loop`` as a background process which
        checks for a sync token; if one exists it sets up the periodic
        ``_process_sync`` call via ``looping_background_call``.
        """
        if self._is_running or not self.is_enabled:
            return
        self._is_running = True
        self.api.run_as_background_process(
            "famedly_control_group_membership_sync", self.start_sync_loop
        )

    async def start_sync_loop(self) -> None:
        """Initialize state and start the periodic sync.

        If the sync token hasn't been loaded yet, tries to read it from the
        database.  If it's still not available (no managed room created yet),
        returns and will be retried on the next api request. Otherwise sets
        up the periodic ``_process_sync``.
        """
        entry = await self.repository.get_sync_token_entry()
        if entry is None:
            self._is_running = False
            return
        self._sync_token_user_id, self._sync_token = entry
        self.api.looping_background_call(
            self._process_sync,
            self.polling_interval_seconds * 1000,
            desc="famedly_control_group_membership_sync",
        )

    async def _process_sync(self) -> None:
        """Execute a single sync iteration: fetch diffs and apply membership changes."""
        response = await self.client.get_all_groups_diffs(
            sync=self._sync_token, timeout=30
        )

        sync_succeeded = True

        if response.data:
            group_rooms = await self.repository.get_rooms_by_group()

            for group_id, diffs in response.data.items():
                rooms = group_rooms.get(group_id, [])
                if not rooms:
                    logger.debug(
                        "Received diff for group %s but no managed rooms use it",
                        group_id,
                    )
                    continue

                for room_id, admin_user_id in rooms:
                    if not await self._apply_diffs(room_id, admin_user_id, diffs):
                        sync_succeeded = False

        if not sync_succeeded:
            logger.warning(
                "Some membership updates failed during sync, will retry next cycle",
            )
            return

        self._sync_token = response.next_sync
        if self._sync_token_user_id is not None:
            await self.repository.set_sync_token(
                self._sync_token_user_id, self._sync_token
            )

    async def _apply_diffs(
        self, room_id: str, admin_user_id: str, diffs: list[DiffRecord]
    ) -> bool:
        """Apply membership diffs to a single room.

        Returns:
            True if all operations succeeded, False if any failed.
        """
        external_ids_to_add = [
            d.external_user_id for d in diffs if d.action == Membership.ADD
        ]
        external_ids_to_remove = [
            d.external_user_id for d in diffs if d.action == Membership.REM
        ]

        result = await self.room_handler.apply_membership_changes_from_external_ids(
            room_id=room_id,
            admin_user_id=admin_user_id,
            external_ids_to_add=external_ids_to_add,
            external_ids_to_remove=external_ids_to_remove,
        )

        if result.has_errors:
            if result.not_found_ids:
                logger.warning(
                    "Could not resolve external IDs for room %s: %s",
                    room_id,
                    result.not_found_ids,
                )
            if result.join_errors:
                logger.warning(
                    "Failed to add some users to room %s: %s",
                    room_id,
                    result.join_errors,
                )
            if result.leave_errors:
                logger.warning(
                    "Failed to remove some users from room %s: %s",
                    room_id,
                    result.leave_errors,
                )
            return False

        return True



import logging
import re
from typing import Iterable, Pattern

from pydantic import ValidationError
from synapse.api.constants import CREATOR_POWER_LEVEL
from synapse.api.room_versions import KNOWN_ROOM_VERSIONS
from synapse.http.servlet import (
    RestServlet,
    parse_integer,
    parse_json_object_from_request,
)
from synapse.http.site import SynapseRequest
from synapse.module_api import ModuleApi
from synapse.types import JsonDict, RoomID

from famedly_control_synapse.client import FamedlyControlClient, FamedlyControlError
from famedly_control_synapse.repository import ManagedRoomRepository
from famedly_control_synapse.rest.types import (
    MANAGED_ROOM_TYPE,
    AssignGroupsToManagedRoomRequest,
    CreateManagedRoomRequest,
)
from famedly_control_synapse.room_handler import ManagedRoomHandler

MANAGED_ROOM_API_PREFIX = "/_famedlyControl/v1/managedRooms"
logger = logging.getLogger(__name__)


def famedly_control_patterns(path_regex: str) -> Iterable[Pattern]:
    """Returns the list of patterns for a managed room endpoint

    Args:
        path_regex: The regex string to match. This should NOT have a ^
            as this will be prefixed.

    Returns:
        A list of regex patterns.
    """
    famedly_control_prefix = "^" + MANAGED_ROOM_API_PREFIX
    patterns = [re.compile(famedly_control_prefix + path_regex)]
    return patterns


class CreateManagedRoomResource(RestServlet):
    """Resource for creating a new managed room."""

    PATTERNS = famedly_control_patterns("/createRoom")

    def __init__(
        self,
        api: ModuleApi,
        client: FamedlyControlClient,
        room_handler: ManagedRoomHandler,
        repository: ManagedRoomRepository,
    ) -> None:
        super().__init__()
        self.api = api
        self.client = client
        self.account_data_handler = self.api._account_data_handler
        self.room_handler = room_handler
        self.repository = repository

    async def on_POST(self, request: SynapseRequest) -> tuple[int, JsonDict]:
        """Handle POST requests to create a new managed room."""
        requester = await self.api.get_user_by_req(request)
        admin_user_id = requester.user.to_string()

        if not await self.api.is_user_admin(admin_user_id):
            return 403, {"error": "user is not administrator"}

        room_config = parse_json_object_from_request(request)

        try:
            if room_config.get("room_version") is None:
                room_config["room_version"] = (
                    self.api._hs.config.server.default_room_version.identifier
                )
            validated_room_config = CreateManagedRoomRequest.model_validate(room_config)
        except ValidationError as e:
            validation_error = [
                {"loc": err.get("loc"), "msg": err.get("msg")} for err in e.errors()
            ]
            return 400, {"error": "Invalid request body", "details": validation_error}

        room_version = KNOWN_ROOM_VERSIONS.get(validated_room_config.room_version)
        if room_version is None:
            return 400, {
                "error": f"Unsupported room version: {validated_room_config.room_version}"
            }

        if not room_version.msc4289_creator_power_enabled:
            if "power_level_content_override" in room_config:
                validated_room_config.power_level_content_override.users = {
                    admin_user_id: CREATOR_POWER_LEVEL - 1
                }

        room_id, _ = await self.api.create_room(
            admin_user_id,
            validated_room_config.model_dump(by_alias=True, exclude_none=True),
        )

        await self.account_data_handler.add_account_data_to_room(
            admin_user_id,
            room_id,
            MANAGED_ROOM_TYPE,
            {"groups": validated_room_config.groups},
        )

        await self.repository.initialize_sync_token(admin_user_id)

        member_external_ids = set()
        try:
            for group_id in validated_room_config.groups:
                members = await self.client.get_group_members(group_id)
                member_external_ids.update(members)
        except FamedlyControlError as e:
            return e.code, {"error": e.msg}

        user_mxids, not_founds = (
            await self.room_handler.batch_convert_external_user_ids_to_matrix_user_ids(
                list(member_external_ids)
            )
        )
        if not_founds:
            return 400, {
                "error": "Some external user IDs could not be mapped to Matrix user IDs",
                "details": not_founds,
            }

        join_errors = await self.room_handler.force_join_users_to_room(
            room_id, user_mxids, requester
        )
        if join_errors:
            logger.warning(
                "Some members failed to join room %s: %s", room_id, join_errors
            )
            return 207, {
                "error": "Failed to add some members to the room",
                "details": join_errors,
            }

        return 200, {"room_id": room_id}


class ListManagedRoomsResource(RestServlet):
    """Resource for listing all managed rooms."""

    PATTERNS = famedly_control_patterns("/rooms")

    def __init__(self, api: ModuleApi, repository: ManagedRoomRepository) -> None:
        super().__init__()
        self.api = api
        self.repo = repository

    async def on_GET(self, request: SynapseRequest) -> tuple[int, JsonDict]:
        """Handle GET requests to list managed rooms."""
        requester = await self.api.get_user_by_req(request)
        user_id = requester.user.to_string()

        if not await self.api.is_user_admin(user_id):
            return 403, {"error": "user is not administrator"}

        # The 'from' query parameter is labeled as a string in the openapi spec, but is
        # passed directly into the sql query which expects it to be an integer(for
        # OFFSET). Just parse it as the integer directly. This allows it to have a
        # default when one is not supplied, disallows negative numbers, and will raise
        # as a 400 with M_INVALID_PARAM when it is not a legitimate integer. Since the
        # 'from' parameter should have come from a previous page of this endpoint, this
        # should be safe.
        from_token = parse_integer(request, "from", default=0)
        limit = parse_integer(request, "limit", default=100)

        total_count = await self.repo.count_managed_rooms()
        entries = await self.repo.get_managed_rooms_paginated(limit + 1, from_token)

        has_next = len(entries) > limit
        chunk = entries[:limit]

        response: JsonDict = {
            "chunk": chunk,
            "total_room_count_estimate": total_count,
        }

        if has_next:
            response["next_batch"] = str(from_token + limit)
        if from_token > 0:
            response["prev_batch"] = str(max(0, from_token - limit))

        return 200, response


class AssignGroupsToManagedRoomResource(RestServlet):
    """Resource for assigning groups to a managed room."""

    PATTERNS = famedly_control_patterns("/(?P<room_id>[^/]*)/groups")

    def __init__(
        self,
        api: ModuleApi,
        client: FamedlyControlClient,
        room_handler: ManagedRoomHandler,
        repository: ManagedRoomRepository,
    ) -> None:
        super().__init__()
        self.api = api
        self.client = client
        self.account_data_handler = self.api._account_data_handler
        self.room_handler = room_handler
        self.repository = repository

    async def on_POST(
        self, request: SynapseRequest, room_id: str
    ) -> tuple[int, JsonDict]:
        """Handle POST requests to assign groups to a managed room."""
        # Validate room_id format to prevent malicious input
        _ = RoomID.from_string(room_id)

        # Validate user permissions
        requester = await self.api.get_user_by_req(request)
        user_id = requester.user.to_string()
        if not await self.api.is_user_admin(user_id):
            return 403, {"error": "user is not administrator"}

        # Validate if it's a managed room
        if not await self.repository.is_managed_room(room_id, user_id):
            return 404, {"error": "Room not found or not a managed room"}

        # Updated groups information from the request body
        try:
            validated_input = AssignGroupsToManagedRoomRequest.model_validate(
                parse_json_object_from_request(request)
            )
        except ValidationError as e:
            validation_error = [
                {"loc": err.get("loc"), "msg": err.get("msg")} for err in e.errors()
            ]
            return 400, {"error": "Invalid request body", "details": validation_error}

        # Get the current members of the room
        current_member_mxids = await self.api._store.get_users_in_room(room_id)
        current_members = set(current_member_mxids)

        # Get the new groups member state (desired state after update)
        try:
            expected_member_external_ids = set()
            for group_id in validated_input.groups:
                members = await self.client.get_group_members(group_id)
                expected_member_external_ids.update(members)
        except FamedlyControlError as e:
            return e.code, {"error": e.msg}

        expected_member_mxids, not_founds = (
            await self.room_handler.batch_convert_external_user_ids_to_matrix_user_ids(
                list(expected_member_external_ids)
            )
        )
        if not_founds:
            return 400, {
                "error": "Some external user IDs could not be mapped to Matrix user IDs",
                "details": not_founds,
            }

        expected_members = set(expected_member_mxids)

        # Calculate membership changes
        members_to_add = expected_members - current_members
        members_to_remove = current_members - expected_members

        # Prevent the room creator/admin from the membership changes.
        room_creator = await self.room_handler.get_room_creator(room_id)
        if room_creator:
            members_to_add.discard(room_creator)
            members_to_remove.discard(room_creator)
        members_to_add.discard(user_id)
        members_to_remove.discard(user_id)

        # Apply membership changes
        result = await self.room_handler.apply_membership_changes(
            room_id, user_id, list(members_to_add), list(members_to_remove)
        )
        if result.leave_errors:
            logger.error(
                "Some members failed to leave room %s: %s",
                room_id,
                result.leave_errors,
            )
        if result.join_errors:
            logger.error(
                "Some members failed to join room %s: %s",
                room_id,
                result.join_errors,
            )

        # Update room account data with new groups information
        await self.update_room_account_data(user_id, room_id, validated_input.groups)

        if result.join_errors and result.leave_errors:
            return 207, {
                "error": "Failed to add and remove some members from the room",
                "details": {
                    "join_errors": result.join_errors,
                    "leave_errors": result.leave_errors,
                },
            }
        elif result.leave_errors:
            return 207, {
                "error": "Failed to remove some members from the room",
                "details": result.leave_errors,
            }
        elif result.join_errors:
            return 207, {
                "error": "Failed to add some members to the room",
                "details": result.join_errors,
            }
        return 200, {"room_id": room_id, "groups": validated_input.groups}

    async def update_room_account_data(
        self, user_id: str, room_id: str, groups: list[str]
    ) -> None:
        """Helper method to update the room account data for a user."""
        try:
            await self.account_data_handler.remove_account_data_for_room(
                user_id, room_id, MANAGED_ROOM_TYPE
            )
            await self.account_data_handler.add_account_data_to_room(
                user_id,
                room_id,
                MANAGED_ROOM_TYPE,
                {"groups": groups},
            )
        except Exception as e:
            logger.exception(
                "Failed to update account data for room %s: %s", room_id, e
            )



import logging
from typing import Final, Literal

from pydantic import BaseModel, Field
from synapse.api.errors import SynapseError
from synapse.module_api import ModuleApi

from famedly_control_synapse.config import FamedlyControlConfig

logger = logging.getLogger(__name__)


class Membership:
    """Action type for group membership changes."""

    ADD: Final = "Add"
    REM: Final = "Rem"


class DiffRecord(BaseModel):
    """One record from a diff for a group."""

    external_user_id: str = Field(alias="user_id")
    action: Literal["Add", "Rem"]  # Add: User was added, Rem: User was removed


class ManyGroupsDiffResponse(BaseModel):
    """Response containing multiple group diffs."""

    next_sync: str
    data: dict[str, list[DiffRecord]]  # Mapping from Group IDs to list of changes


class MemberInfo(BaseModel):
    """External User Id."""

    external_user_id: str = Field(alias="user_id")


class GroupMembersResponse(BaseModel):
    """Response containing list of group members."""

    members: list[MemberInfo]


class FamedlyControlClient:
    def __init__(self, api: ModuleApi, config: FamedlyControlConfig):
        self.access_token = config.famedly_control.access_token
        self.url = config.famedly_control.api_url
        self.http_client = api.http_client
        self.sync = 0

    async def get_group_members(self, group_id: str) -> list[str]:
        """Get the current members of a group.

        Args:
            group_id: The UUID of the group.

        Returns:
            List of external user IDs who are members of the group.

        Raises:
            FamedlyControlError: If the API returns an error or network failure occurs.
        """
        uri = str(self.url) + "/get_group_members"
        body = {
            "group_id": group_id,
        }
        try:
            response = await self.http_client.post_json_get_json(
                uri,
                body,
                headers={"Authorization": f"Bearer {self.access_token}"},
            )
            if "Ok" in response:
                validated_response = GroupMembersResponse.model_validate(response["Ok"])
                return [
                    member.external_user_id for member in validated_response.members
                ]
            elif "Err" in response:
                error_type = response["Err"].get("type")
                logger.error("Famedly Control API Error: %s", error_type)
                raise FamedlyControlError(
                    500, f"Famedly Control API Error: {error_type}"
                )
            else:
                logger.error("Famedly Control API Error: %s", response)
                raise FamedlyControlError(
                    500, f"Unexpected response format: {response}"
                )
        except FamedlyControlError:
            raise
        except Exception as e:
            logger.error("Famedly Control API Error: %s", e)
            raise FamedlyControlError(500, f"Famedly Control API Error: {e}") from e

    async def get_all_groups_diffs(
        self, sync: str | None, timeout: int = 30
    ) -> ManyGroupsDiffResponse:
        """Get group membership diffs for all known groups. Long polling.

        Warning: This can return a lot of data!

        Args:
            sync: Monotonically increasing sync token from previous response.
            timeout: How long to wait in seconds if response would be empty.

        Returns:
            ManyGroupsDiffResponse containing next_sync token and mapping of group diffs.

        Raises:
            FamedlyControlError: If the API returns an error or network failure occurs.
        """
        uri = str(self.url) + "/get_all_groups_diffs"
        body: dict = {
            "timeout": timeout,
        }
        if sync is not None:
            body["sync"] = sync
        try:
            response = await self.http_client.post_json_get_json(
                uri,
                body,
                headers={"Authorization": f"Bearer {self.access_token}"},
            )
            if "Ok" in response:
                return ManyGroupsDiffResponse.model_validate(response["Ok"])
            elif "Err" in response:
                error_type = response["Err"].get("type")
                logger.error("Famedly Control API Error: %s", error_type)
                raise FamedlyControlError(
                    500, f"Famedly Control API Error: {error_type}"
                )
            else:
                logger.error("Famedly Control API Error: %s", response)
                raise FamedlyControlError(
                    500, f"Unexpected response format: {response}"
                )
        except FamedlyControlError:
            raise
        except Exception as e:
            logger.error("Famedly Control API Error: %s", e)
            raise FamedlyControlError(500, f"Famedly Control API Error: {e}") from e


class FamedlyControlError(SynapseError):
    """Base exception for FamedlyControl API errors."""

    code = 500
    msg = "An error occurred with the Famedly Control API"

    def __init__(self, code: int | None = None, msg: str | None = None):
        self.msg = msg or self.__class__.msg
        self.code = code or self.__class__.code
        super().__init__(self.code, self.msg)



# Copyright 2014-2016 OpenMarket Ltd
# Copyright 2017 Vector Creations Ltd
# Copyright 2018-2019 New Vector Ltd
# Copyright 2019-2021 The Matrix.org Foundation C.I.C.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import time
from collections.abc import Iterable
from http import HTTPStatus
from typing import Any, AnyStr, Literal, overload
from urllib.parse import urlencode

import attr
from synapse.api.constants import Membership
from synapse.api.errors import Codes
from synapse.server import HomeServer
from synapse.types import JsonDict
from twisted.internet.testing import MemoryReactorClock
from twisted.web.server import Site

from tests.utils.server import make_request

# an 'oidc_config' suitable for login_via_oidc.
TEST_OIDC_ISSUER = "https://issuer.test/"
TEST_OIDC_CONFIG = {
    "enabled": True,
    "issuer": TEST_OIDC_ISSUER,
    "client_id": "test-client-id",
    "client_secret": "test-client-secret",
    "scopes": ["openid"],
    "user_mapping_provider": {"config": {"localpart_template": "{{ user.sub }}"}},
}


@attr.s(auto_attribs=True)
class RestHelper:
    """Contains extra helper functions to quickly and clearly perform a given
    REST action, which isn't the focus of the test.
    """

    hs: HomeServer
    reactor: MemoryReactorClock
    site: Site
    auth_user_id: str | None

    @overload
    def create_room_as(
        self,
        room_creator: str | None = ...,
        is_public: bool | None = ...,
        room_version: str | None = ...,
        tok: str | None = ...,
        expect_code: Literal[200] = ...,
        extra_content: dict | None = ...,
        custom_headers: Iterable[tuple[AnyStr, AnyStr]] | None = ...,
    ) -> str: ...

    @overload
    def create_room_as(
        self,
        room_creator: str | None = ...,
        is_public: bool | None = ...,
        room_version: str | None = ...,
        tok: str | None = ...,
        expect_code: int = ...,
        extra_content: dict | None = ...,
        custom_headers: Iterable[tuple[AnyStr, AnyStr]] | None = ...,
    ) -> str | None: ...

    def create_room_as(
        self,
        room_creator: str | None = None,
        is_public: bool | None = True,
        room_version: str | None = None,
        tok: str | None = None,
        expect_code: int = HTTPStatus.OK,
        extra_content: dict | None = None,
        custom_headers: Iterable[tuple[AnyStr, AnyStr]] | None = None,
    ) -> str | None:
        """
        Create a room.

        Args:
            room_creator: The user ID to create the room with.
            is_public: If True, the `visibility` parameter will be set to
                "public". If False, it will be set to "private".
                If None, doesn't specify the `visibility` parameter in which
                case the server is supposed to make the room private according to
                the CS API.
                Defaults to public, since that is commonly needed in tests
                for convenience where room privacy is not a problem.
            room_version: The room version to create the room as. Defaults to Synapse's
                default room version.
            tok: The access token to use in the request.
            expect_code: The expected HTTP response code.
            extra_content: Extra keys to include in the body of the /createRoom request.
                Note that if is_public is set, the "visibility" key will be overridden.
                If room_version is set, the "room_version" key will be overridden.
            custom_headers: HTTP headers to include in the request.

        Returns:
            The ID of the newly created room, or None if the request failed.
        """
        temp_id = self.auth_user_id
        self.auth_user_id = room_creator
        path = "/_matrix/client/r0/createRoom"
        content = extra_content or {}
        if is_public is not None:
            content["visibility"] = "public" if is_public else "private"
        if room_version:
            content["room_version"] = room_version
        if tok:
            path = f"{path}?access_token={tok}"

        channel = make_request(
            self.reactor,
            self.site,
            "POST",
            path,
            content,
            custom_headers=custom_headers,
        )

        assert channel.code == expect_code, channel.result
        self.auth_user_id = temp_id

        if expect_code == HTTPStatus.OK:
            return channel.json_body["room_id"]

        return None

    def invite(
        self,
        room: str,
        src: str | None = None,
        targ: str | None = None,
        expect_code: int = HTTPStatus.OK,
        tok: str | None = None,
    ) -> None:
        self.change_membership(
            room=room,
            src=src,
            targ=targ,
            tok=tok,
            membership=Membership.INVITE,
            expect_code=expect_code,
        )

    def join(
        self,
        room: str,
        user: str | None = None,
        expect_code: int = HTTPStatus.OK,
        tok: str | None = None,
        appservice_user_id: str | None = None,
        expect_errcode: Codes | None = None,
        expect_additional_fields: dict | None = None,
    ) -> None:
        self.change_membership(
            room=room,
            src=user,
            targ=user,
            tok=tok,
            appservice_user_id=appservice_user_id,
            membership=Membership.JOIN,
            expect_code=expect_code,
            expect_errcode=expect_errcode,
            expect_additional_fields=expect_additional_fields,
        )

    def knock(
        self,
        room: str | None = None,
        user: str | None = None,
        reason: str | None = None,
        expect_code: int = HTTPStatus.OK,
        tok: str | None = None,
    ) -> None:
        temp_id = self.auth_user_id
        self.auth_user_id = user
        path = f"/knock/{room}"
        if tok:
            path = f"{path}?access_token={tok}"

        data = {}
        if reason:
            data["reason"] = reason

        channel = make_request(
            self.reactor,
            self.site,
            "POST",
            path,
            data,
        )

        assert channel.code == expect_code, "Expected: %d, got: %d, resp: %r" % (
            expect_code,
            channel.code,
            channel.result["body"],
        )

        self.auth_user_id = temp_id

    def leave(
        self,
        room: str,
        user: str | None = None,
        expect_code: int = HTTPStatus.OK,
        tok: str | None = None,
    ) -> None:
        self.change_membership(
            room=room,
            src=user,
            targ=user,
            tok=tok,
            membership=Membership.LEAVE,
            expect_code=expect_code,
        )

    def ban(
        self,
        room: str,
        src: str,
        targ: str,
        expect_code: int = HTTPStatus.OK,
        tok: str | None = None,
    ) -> None:
        """A convenience helper: `change_membership` with `membership` preset to "ban"."""
        self.change_membership(
            room=room,
            src=src,
            targ=targ,
            tok=tok,
            membership=Membership.BAN,
            expect_code=expect_code,
        )

    def change_membership(
        self,
        room: str,
        src: str | None,
        targ: str | None,
        membership: str,
        extra_data: dict | None = None,
        tok: str | None = None,
        appservice_user_id: str | None = None,
        expect_code: int = HTTPStatus.OK,
        expect_errcode: str | None = None,
        expect_additional_fields: dict | None = None,
    ) -> None:
        """
        Send a membership state event into a room.

        Args:
            room: The ID of the room to send to
            src: The mxid of the event sender
            targ: The mxid of the event's target. The state key
            membership: The type of membership event
            extra_data: Extra information to include in the content of the event
            tok: The user access token to use
            appservice_user_id: The `user_id` URL parameter to pass.
                This allows driving an application service user
                using an application service access token in `tok`.
            expect_code: The expected HTTP response code
            expect_errcode: The expected Matrix error code
        """
        temp_id = self.auth_user_id
        self.auth_user_id = src

        path = f"/_matrix/client/r0/rooms/{room}/state/m.room.member/{targ}"
        url_params: dict[str, str] = {}

        if tok:
            url_params["access_token"] = tok

        if appservice_user_id:
            url_params["user_id"] = appservice_user_id

        if url_params:
            path += "?" + urlencode(url_params)

        data = {"membership": membership}
        data.update(extra_data or {})

        channel = make_request(
            self.reactor,
            self.site,
            "PUT",
            path,
            data,
        )

        assert channel.code == expect_code, "Expected: %d, got: %d, resp: %r" % (
            expect_code,
            channel.code,
            channel.result["body"],
        )

        if expect_errcode:
            assert (
                str(channel.json_body["errcode"]) == expect_errcode
            ), "Expected: {!r}, got: {!r}, resp: {!r}".format(
                expect_errcode,
                channel.json_body["errcode"],
                channel.result["body"],
            )

        if expect_additional_fields is not None:
            for expect_key, expect_value in expect_additional_fields.items():
                assert (
                    expect_key in channel.json_body
                ), f"Expected field {expect_key}, got {channel.json_body}"
                assert (
                    channel.json_body[expect_key] == expect_value
                ), f"Expected: {expect_value} at {expect_key}, got: {channel.json_body[expect_key]}, resp: {channel.json_body}"

        self.auth_user_id = temp_id

    def send(
        self,
        room_id: str,
        body: str | None = None,
        txn_id: str | None = None,
        tok: str | None = None,
        expect_code: int = HTTPStatus.OK,
        custom_headers: Iterable[tuple[AnyStr, AnyStr]] | None = None,
    ) -> JsonDict:
        if body is None:
            body = "body_text_here"

        content = {"msgtype": "m.text", "body": body}

        return self.send_event(
            room_id,
            "m.room.message",
            content,
            txn_id,
            tok,
            expect_code,
            custom_headers=custom_headers,
        )

    def send_event(
        self,
        room_id: str,
        type_: str,
        content: dict | None = None,
        txn_id: str | None = None,
        tok: str | None = None,
        expect_code: int = HTTPStatus.OK,
        custom_headers: Iterable[tuple[AnyStr, AnyStr]] | None = None,
    ) -> JsonDict:
        if txn_id is None:
            txn_id = f"m{time.time()!s}"

        path = f"/_matrix/client/r0/rooms/{room_id}/send/{type_}/{txn_id}"
        if tok:
            path = f"{path}?access_token={tok}"

        channel = make_request(
            self.reactor,
            self.site,
            "PUT",
            path,
            content or {},
            custom_headers=custom_headers,
        )

        assert channel.code == expect_code, "Expected: %d, got: %d, resp: %r" % (
            expect_code,
            channel.code,
            channel.result["body"],
        )

        return channel.json_body

    def get_event(
        self,
        room_id: str,
        event_id: str,
        tok: str | None = None,
        expect_code: int = HTTPStatus.OK,
    ) -> JsonDict:
        """Request a specific event from the server.

        Args:
            room_id: the room in which the event was sent.
            event_id: the event's ID.
            tok: the token to request the event with.
            expect_code: the expected HTTP status for the response.

        Returns:
            The event as a dict.
        """
        path = f"/_matrix/client/v3/rooms/{room_id}/event/{event_id}"
        if tok:
            path = path + f"?access_token={tok}"

        channel = make_request(
            self.reactor,
            self.site,
            "GET",
            path,
        )

        assert channel.code == expect_code, "Expected: %d, got: %d, resp: %r" % (
            expect_code,
            channel.code,
            channel.result["body"],
        )

        return channel.json_body

    def _read_write_state(
        self,
        room_id: str,
        event_type: str,
        body: dict[str, Any] | None,
        tok: str | None,
        expect_code: int = HTTPStatus.OK,
        state_key: str = "",
        method: str = "GET",
    ) -> JsonDict:
        """Read or write some state from a given room

        Args:
            room_id:
            event_type: The type of state event
            body: Body that is sent when making the request. The content of the state event.
                If None, the request to the server will have an empty body
            tok: The access token to use
            expect_code: The HTTP code to expect in the response
            state_key:
            method: "GET" or "PUT" for reading or writing state, respectively

        Returns:
            The response body from the server

        Raises:
            AssertionError: if expect_code doesn't match the HTTP code we received
        """
        path = f"/_matrix/client/r0/rooms/{room_id}/state/{event_type}/{state_key}"
        if tok:
            path = f"{path}?access_token={tok}"

        # Set request body if provided
        content = b""
        if body is not None:
            content = json.dumps(body).encode("utf8")

        channel = make_request(self.reactor, self.site, method, path, content)

        assert (
            channel.code == expect_code
        ), f"Expected: {expect_code}, got: {channel.code}, resp: {channel.result['body']}"

        return channel.json_body

    def get_state(
        self,
        room_id: str,
        event_type: str,
        tok: str,
        expect_code: int = HTTPStatus.OK,
        state_key: str = "",
    ) -> JsonDict:
        """Gets some state from a room

        Args:
            room_id:
            event_type: The type of state event
            tok: The access token to use
            expect_code: The HTTP code to expect in the response
            state_key:

        Returns:
            The response body from the server

        Raises:
            AssertionError: if expect_code doesn't match the HTTP code we received
        """
        return self._read_write_state(
            room_id, event_type, None, tok, expect_code, state_key, method="GET"
        )

    def send_state(
        self,
        room_id: str,
        event_type: str,
        body: dict[str, Any],
        tok: str | None,
        expect_code: int = HTTPStatus.OK,
        state_key: str = "",
    ) -> JsonDict:
        """Set some state in a room

        Args:
            room_id:
            event_type: The type of state event
            body: Body that is sent when making the request. The content of the state event.
            tok: The access token to use
            expect_code: The HTTP code to expect in the response
            state_key:

        Returns:
            The response body from the server

        Raises:
            AssertionError: if expect_code doesn't match the HTTP code we received
        """
        return self._read_write_state(
            room_id, event_type, body, tok, expect_code, state_key, method="PUT"
        )



# Copyright 2014-2016 OpenMarket Ltd
# Copyright 2018 New Vector
# Copyright 2019 Matrix.org Federation C.I.C
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
import gc
import hashlib
import hmac
import json
import logging
import secrets
import time
from collections.abc import Awaitable, Callable, Iterable, Mapping
from http import HTTPStatus
from typing import Any, ClassVar, Concatenate, Generic, NoReturn, Protocol, TypeVar
from unittest.mock import Mock, patch

from synapse import events
from synapse.api.constants import EventTypes
from synapse.config._base import Config, RootConfig
from synapse.config.homeserver import HomeServerConfig
from synapse.http.server import JsonResource, OptionsResource
from synapse.http.site import SynapseRequest, SynapseSite
from synapse.logging.context import (
    SENTINEL_CONTEXT,
    LoggingContext,
    current_context,
    set_current_context,
)
from synapse.rest import RegisterServletsFunc
from synapse.server import HomeServer
from synapse.types import JsonDict, Requester, UserID, create_requester
from synapse.util.clock import Clock
from synapse.util.httpresourcetree import create_resource_tree
from twisted.internet.defer import Deferred, ensureDeferred
from twisted.internet.testing import MemoryReactor, MemoryReactorClock
from twisted.python.failure import Failure
from twisted.python.threadpool import ThreadPool
from twisted.trial import unittest
from twisted.web.resource import Resource
from twisted.web.server import Request
from typing_extensions import ParamSpec

from tests.utils.config import (
    checked_cast,
    default_config,
    setup_awaitable_errors,
    setupdb,
)
from tests.utils.logging import setup_logging
from tests.utils.rest_helper import RestHelper
from tests.utils.server import (
    CustomHeaderType,
    FakeChannel,
    ThreadedMemoryReactorClock,
    get_clock,
    make_request,
    setup_test_homeserver,
)

setupdb()
setup_logging()

TV = TypeVar("TV")
_ExcType_co = TypeVar("_ExcType_co", bound=BaseException, covariant=True)

P = ParamSpec("P")
R = TypeVar("R")
S = TypeVar("S")


class _TypedFailure(Generic[_ExcType_co], Protocol):
    """Extension to twisted.Failure, where the 'value' has a certain type."""

    @property
    def value(self) -> _ExcType_co: ...


def around(target: TV) -> Callable[[Callable[Concatenate[S, P], R]], None]:
    """A CLOS-style 'around' modifier, which wraps the original method of the
    given instance with another piece of code.

    @around(self)
    def method_name(orig, *args, **kwargs):
        return orig(*args, **kwargs)
    """

    def _around(code: Callable[Concatenate[S, P], R]) -> None:
        name = code.__name__
        orig = getattr(target, name)

        def new(*args: P.args, **kwargs: P.kwargs) -> R:
            return code(orig, *args, **kwargs)

        setattr(target, name, new)

    return _around


_TConfig = TypeVar("_TConfig", Config, RootConfig)


def deepcopy_config(config: _TConfig) -> _TConfig:
    new_config: _TConfig

    if isinstance(config, RootConfig):
        new_config = config.__class__(config.config_files)  # type: ignore[arg-type]
    else:
        new_config = config.__class__(config.root)

    for attr_name in config.__dict__:
        if attr_name.startswith("__") or attr_name == "root":
            continue
        attr = getattr(config, attr_name)
        new_attr = deepcopy_config(attr) if isinstance(attr, Config) else attr

        setattr(new_config, attr_name, new_attr)

    return new_config


@functools.lru_cache(maxsize=8)
def _parse_config_dict(config: str) -> RootConfig:
    config_obj = HomeServerConfig()
    config_obj.parse_config_dict(json.loads(config), "", "")
    return config_obj


def make_homeserver_config_obj(config: dict[str, Any]) -> RootConfig:
    """Creates a :class:`HomeServerConfig` instance with the given configuration dict.

    This is equivalent to::

        config_obj = HomeServerConfig()
        config_obj.parse_config_dict(config, "", "")

    but it keeps a cache of `HomeServerConfig` instances and deepcopies them as needed,
    to avoid validating the whole configuration every time.
    """
    config_obj = _parse_config_dict(json.dumps(config, sort_keys=True))
    return deepcopy_config(config_obj)


class TestCase(unittest.TestCase):
    """A subclass of twisted.trial's TestCase which looks for 'loglevel'
    attributes on both itself and its individual test methods, to override the
    root logger's logging level while that test (case|method) runs."""

    def __init__(self, methodName: str):
        super().__init__(methodName)

        method = getattr(self, methodName)

        level = getattr(method, "loglevel", getattr(self, "loglevel", None))

        @around(self)
        def setUp(orig: Callable[[], R]) -> R:
            # if we're not starting in the sentinel logcontext, then to be honest
            # all future bets are off.
            if current_context():
                self.fail(
                    f"Test starting with non-sentinel logging context {current_context()}"
                )

            # Disable GC for duration of test. See below for why.
            gc.disable()

            old_level = logging.getLogger().level
            if level is not None and old_level != level:

                @around(self)
                def tearDown(orig: Callable[[], R]) -> R:
                    ret = orig()
                    logging.getLogger().setLevel(old_level)
                    return ret

                logging.getLogger().setLevel(level)

            # Trial messes with the warnings configuration, thus this has to be
            # done in the context of an individual TestCase.
            self.addCleanup(setup_awaitable_errors())

            return orig()

        # We want to force a GC to workaround problems with deferreds leaking
        # logcontexts when they are GCed (see the logcontext docs).
        #
        # The easiest way to do this would be to do a full GC after each test
        # run, but that is very expensive. Instead, we disable GC (above) for
        # the duration of the test and only run a gen-0 GC, which is a lot
        # quicker. This doesn't clean up everything, since the TestCase
        # instance still holds references to objects created during the test,
        # such as HomeServers, so we do a full GC every so often.

        @around(self)
        def tearDown(orig: Callable[[], R]) -> R:
            ret = orig()
            gc.collect(0)
            # Run a full GC every 50 gen-0 GCs.
            gen0_stats = gc.get_stats()[0]
            gen0_collections = gen0_stats["collections"]
            if gen0_collections % 50 == 0:
                gc.collect()
            gc.enable()
            set_current_context(SENTINEL_CONTEXT)

            return ret

    def assertObjectHasAttributes(self, attrs: dict[str, object], obj: object) -> None:
        """Asserts that the given object has each of the attributes given, and
        that the value of each matches according to assertEqual."""
        for key in attrs:
            if not hasattr(obj, key):
                msg = f"Expected obj to have a '.{key}'"
                raise AssertionError(msg)
            try:
                assert attrs[key] == getattr(obj, key)
            except AssertionError as e:
                msg = f"Assert error for '.{key}':"
                raise (type(e))(msg) from e

    def assert_dict(self, required: Mapping, actual: Mapping) -> None:
        """Does a partial assert of a dict.

        Args:
            required: The keys and value which MUST be in 'actual'.
            actual: The test result. Extra keys will not be checked.
        """
        for key in required:
            assert required[key] == actual[key], f"{key} mismatch. {actual}"


def DEBUG(target: TV) -> TV:
    """A decorator to set the .loglevel attribute to logging.DEBUG.
    Can apply to either a TestCase or an individual test method."""
    target.loglevel = logging.DEBUG  # type: ignore[attr-defined]
    return target


def INFO(target: TV) -> TV:
    """A decorator to set the .loglevel attribute to logging.INFO.
    Can apply to either a TestCase or an individual test method."""
    target.loglevel = logging.INFO  # type: ignore[attr-defined]
    return target


def logcontext_clean(target: TV) -> TV:
    """A decorator which marks the TestCase or method as 'logcontext_clean'

    ... ie, any logcontext errors should cause a test failure
    """

    def logcontext_error(msg: str) -> NoReturn:
        err = f"logcontext error: {msg}"
        raise AssertionError(err)

    patcher = patch("synapse.logging.context.logcontext_error", new=logcontext_error)
    return patcher(target)  # type: ignore[call-overload]


class HomeserverTestCase(TestCase):
    """
    A base TestCase that reduces boilerplate for HomeServer-using test cases.

    Defines a setUp method which creates a mock reactor, and instantiates a homeserver
    running on that reactor.

    There are various hooks for modifying the way that the homeserver is instantiated:

    * override make_homeserver, for example by making it pass different parameters into
      setup_test_homeserver.

    * override default_config, to return a modified configuration dictionary for use
      by setup_test_homeserver.

    * On a per-test basis, you can use the @override_config decorator to give a
      dictionary containing additional configuration settings to be added to the basic
      config dict.

    Attributes:
        servlets: List of servlet registration function.
        user_id (str): The user ID to assume if auth is hijacked.
        hijack_auth: Whether to hijack auth to return the user specified
           in user_id.
    """

    hijack_auth: ClassVar[bool] = True
    needs_threadpool: ClassVar[bool] = False
    servlets: ClassVar[list[RegisterServletsFunc]] = []

    def __init__(self, methodName: str):
        super().__init__(methodName)

        # see if we have any additional config for this test
        method = getattr(self, methodName)
        self._extra_config = getattr(method, "_extra_config", None)

    def setUp(self) -> None:
        """
        Set up the TestCase by calling the homeserver constructor, optionally
        hijacking the authentication system to return a fixed user, and then
        calling the prepare function.
        """
        self.reactor, self.clock = get_clock()
        self._hs_args = {"clock": self.clock, "reactor": self.reactor}
        self.hs = self.make_homeserver(self.reactor, self.clock)

        # Honour the `use_frozen_dicts` config option. We have to do this
        # manually because this is taken care of in the app `start` code, which
        # we don't run. Plus we want to reset it on tearDown.
        events.USE_FROZEN_DICTS = self.hs.config.server.use_frozen_dicts

        if self.hs is None:
            msg = "No homeserver returned from make_homeserver."
            raise TypeError(msg)

        if not isinstance(self.hs, HomeServer):
            msg = f"A homeserver wasn't returned, but {self.hs!r}"
            raise TypeError(msg)

        # create the root resource, and a site to wrap it.
        self.resource = self.create_test_resource()
        self.site = SynapseSite(
            logger_name="synapse.access.http.fake",
            site_tag=self.hs.config.server.server_name,
            config=self.hs.config.server.listeners[0],
            resource=self.resource,
            server_version_string="1",
            max_request_body_size=4096,
            reactor=self.reactor,
            hs=self.hs,
        )

        self.helper = RestHelper(  # type: ignore[call-arg]
            self.hs,
            checked_cast(MemoryReactorClock, self.hs.get_reactor()),
            self.site,
            getattr(self, "user_id", None),
        )

        if hasattr(self, "user_id") and self.hijack_auth:
            assert self.helper.auth_user_id is not None
            token = "some_fake_token"

            # We need a valid token ID to satisfy foreign key constraints.
            token_id = self.get_success(
                self.hs.get_datastores().main.add_access_token_to_user(
                    self.helper.auth_user_id,
                    token,
                    None,
                    None,
                )
            )

            # This has to be a function and not just a Mock, because
            # `self.helper.auth_user_id` is temporarily reassigned in some tests
            async def get_requester(*_args: Any, **_kwargs: Any) -> Requester:
                assert self.helper.auth_user_id is not None
                return create_requester(
                    user_id=UserID.from_string(self.helper.auth_user_id),
                    access_token_id=token_id,
                )

            # Type ignore: mypy doesn't like us assigning to methods.
            self.hs.get_auth().get_user_by_req = get_requester  # type: ignore[method-assign]
            self.hs.get_auth().get_user_by_access_token = get_requester  # type: ignore[method-assign]
            self.hs.get_auth().get_access_token_from_request = Mock(return_value=token)  # type: ignore[method-assign]

        if self.needs_threadpool:
            self.reactor.threadpool = ThreadPool()  # type: ignore[assignment]
            self.addCleanup(self.reactor.threadpool.stop)
            self.reactor.threadpool.start()

        if hasattr(self, "prepare"):
            self.prepare(self.reactor, self.clock, self.hs)

    def tearDown(self) -> None:
        # Reset to not use frozen dicts.
        events.USE_FROZEN_DICTS = False
        # Make sure to close the database connection, or we get ResourceWarnings
        self.hs.get_datastores().main.db_pool._db_pool.close()

    def wait_on_thread(self, deferred: Deferred, timeout: int = 10) -> None:
        """
        Wait until a Deferred is done, where it's waiting on a real thread.
        """
        start_time = time.time()

        while not deferred.called:
            if start_time + timeout < time.time():
                msg = "Timed out waiting for threadpool"
                raise ValueError(msg)
            self.reactor.advance(0.01)
            time.sleep(0.01)

    def wait_for_background_updates(self) -> None:
        """Block until all background database updates have completed."""
        store = self.hs.get_datastores().main
        while not self.get_success(
            store.db_pool.updates.has_completed_background_updates()
        ):
            self.get_success(
                store.db_pool.updates.do_next_background_update(False), by=0.1
            )

    def make_homeserver(
        self, reactor: ThreadedMemoryReactorClock, clock: Clock
    ) -> HomeServer:
        """
        Make and return a homeserver.

        Args:
            reactor: A Twisted Reactor, or something that pretends to be one.
            clock: The Clock, associated with the reactor.

        Returns:
            A homeserver suitable for testing.

        Function to be overridden in subclasses.
        """
        return self.setup_test_homeserver(reactor=reactor, clock=clock)

    def create_test_resource(self) -> Resource:
        """
        Create the root resource for the test server.

        The default calls `self.create_resource_dict` and builds the resultant dict
        into a tree.
        """
        root_resource = OptionsResource()
        create_resource_tree(self.create_resource_dict(), root_resource)
        return root_resource

    def create_resource_dict(self) -> dict[str, Resource]:
        """Create a resource tree for the test server

        A resource tree is a mapping from path to twisted.web.resource.

        The default implementation creates a JsonResource and calls each function in
        `servlets` to register servlets against it.
        """
        servlet_resource = JsonResource(self.hs)
        for servlet in self.servlets:
            servlet(self.hs, servlet_resource)
        resources: dict[str, Resource] = {
            "/_matrix/client": servlet_resource,
            "/_synapse/admin": servlet_resource,
        }
        resources.update(self.hs._module_web_resources)
        return resources

    def default_config(self) -> JsonDict:
        """
        Get a default HomeServer config dict.
        """
        config = default_config("test")

        # apply any additional config which was specified via the override_config
        # decorator.
        if self._extra_config is not None:
            config.update(self._extra_config)

        return config

    def prepare(
        self, reactor: MemoryReactor, clock: Clock, homeserver: HomeServer
    ) -> None:
        """
        Prepare for the test.  This involves things like mocking out parts of
        the homeserver, or building test data common across the whole test
        suite.

        Args:
            reactor: A Twisted Reactor, or something that pretends to be one.
            clock: The Clock, associated with the reactor.
            homeserver: The HomeServer to test against.

        Function to optionally be overridden in subclasses.
        """

    def make_request(
        self,
        method: bytes | str,
        path: bytes | str,
        content: bytes | str | JsonDict = b"",
        access_token: str | None = None,
        request: type[Request] = SynapseRequest,
        shorthand: bool = True,
        federation_auth_origin: bytes | None = None,
        content_is_form: bool = False,
        await_result: bool = True,
        custom_headers: Iterable[CustomHeaderType] | None = None,
        client_ip: str = "127.0.0.1",
    ) -> FakeChannel:
        """
        Create a SynapseRequest at the path using the method and containing the
        given content.

        Args:
            method: The HTTP request method ("verb").
            path: The HTTP path, suitably URL encoded (e.g. escaped UTF-8 & spaces
                and such). content (bytes or dict): The body of the request.
                JSON-encoded, if a dict.
            shorthand: Whether to try and be helpful and prefix the given URL
            with the usual REST API path, if it doesn't contain it.
            federation_auth_origin: if set to not-None, we will add a fake
                Authorization header pretending to be the given server name.
            content_is_form: Whether the content is URL encoded form data. Adds the
                'Content-Type': 'application/x-www-form-urlencoded' header.

            await_result: whether to wait for the request to complete rendering. If
                 true (the default), will pump the test reactor until the the renderer
                 tells the channel the request is finished.

            custom_headers: (name, value) pairs to add as request headers

            client_ip: The IP to use as the requesting IP. Useful for testing
                ratelimiting.

        Returns:
            The FakeChannel object which stores the result of the request.
        """
        return make_request(
            self.reactor,
            self.site,
            method,
            path,
            content,
            access_token,
            request,
            shorthand,
            federation_auth_origin,
            content_is_form,
            await_result,
            custom_headers,
            client_ip,
        )

    def setup_test_homeserver(
        self, name: str | None = None, **kwargs: Any
    ) -> HomeServer:
        """
        Set up the test homeserver, meant to be called by the overridable
        make_homeserver. It automatically passes through the test class's
        clock & reactor.

        Args:
            See tests.utils.setup_test_homeserver.

        Returns:
            synapse.server.HomeServer
        """
        kwargs = dict(kwargs)
        kwargs.update(self._hs_args)
        config = self.default_config() if "config" not in kwargs else kwargs["config"]

        # The server name can be specified using either the `name` argument or a config
        # override. The `name` argument takes precedence over any config overrides.
        if name is not None:
            config["server_name"] = name

        # Parse the config from a config dict into a HomeServerConfig
        config_obj = make_homeserver_config_obj(config)
        kwargs["config"] = config_obj

        # The server name in the config is now `name`, if provided, or the `server_name`
        # from a config override, or the default of "test". Whichever it is, we
        # construct a homeserver with a matching name.
        kwargs["name"] = config_obj.server.server_name

        async def run_bg_updates() -> None:
            with LoggingContext(name="run_bg_updates", server_name=kwargs["name"]):
                self.get_success(stor.db_pool.updates.run_background_updates(False))

        hs = setup_test_homeserver(**kwargs)
        stor = hs.get_datastores().main

        # Run the database background updates, when running against "master".
        if hs.__class__.__name__ == "TestHomeServer":
            self.get_success(run_bg_updates())

        return hs

    def pump(self, by: float = 0.0) -> None:
        """
        Pump the reactor enough that Deferreds will fire.
        """
        self.reactor.pump([by] * 100)

    def get_success(self, d: Awaitable[TV], by: float = 0.0) -> TV:
        deferred: Deferred[TV] = ensureDeferred(d)  # type: ignore[arg-type]
        self.pump(by=by)
        return self.successResultOf(deferred)

    def get_failure(
        self, d: Awaitable[Any], exc: type[_ExcType_co]
    ) -> _TypedFailure[_ExcType_co]:
        """
        Run a Deferred and get a Failure from it. The failure must be of the type `exc`.
        """
        deferred: Deferred[Any] = ensureDeferred(d)  # type: ignore[arg-type]
        self.pump()
        return self.failureResultOf(deferred, exc)

    def get_success_or_raise(self, d: Awaitable[TV], by: float = 0.0) -> TV:
        """Drive deferred to completion and return result or raise exception
        on failure.
        """
        deferred: Deferred[TV] = ensureDeferred(d)  # type: ignore[arg-type]

        results: list = []
        deferred.addBoth(results.append)

        self.pump(by=by)

        if not results:
            self.fail(
                f"Success result expected on {deferred!r}, found no result instead"
            )

        result = results[0]

        if isinstance(result, Failure):
            result.raiseException()

        return result

    def register_user(
        self,
        username: str,
        password: str,
        admin: bool | None = False,
        displayname: str | None = None,
    ) -> str:
        """
        Register a user. Requires the Admin API be registered.

        Args:
            username: The user part of the new user.
            password: The password of the new user.
            admin: Whether the user should be created as an admin or not.
            displayname: The displayname of the new user.

        Returns:
            The MXID of the new user.
        """
        self.hs.config.registration.registration_shared_secret = "shared"

        # Create the user
        channel = self.make_request("GET", "/_synapse/admin/v1/register")
        assert channel.code == HTTPStatus.OK, channel.result
        nonce = channel.json_body["nonce"]

        want_mac = hmac.new(key=b"shared", digestmod=hashlib.sha1)
        nonce_str = b"\x00".join([username.encode("utf8"), password.encode("utf8")])
        if admin:
            nonce_str += b"\x00admin"
        else:
            nonce_str += b"\x00notadmin"

        want_mac.update(nonce.encode("ascii") + b"\x00" + nonce_str)
        want_mac_digest = want_mac.hexdigest()

        body = {
            "nonce": nonce,
            "username": username,
            "displayname": displayname,
            "password": password,
            "admin": admin,
            "mac": want_mac_digest,
            "inhibit_login": True,
        }
        channel = self.make_request("POST", "/_synapse/admin/v1/register", body)
        assert channel.code == HTTPStatus.OK, channel.json_body

        return channel.json_body["user_id"]

    def register_appservice_user(
        self,
        username: str,
        appservice_token: str,
    ) -> tuple[str, str]:
        """Register an appservice user as an application service.
        Requires the client-facing registration API be registered.

        Args:
            username: the user to be registered by an application service.
                Should NOT be a full username, i.e. just "localpart" as opposed to "@localpart:hostname"
            appservice_token: the acccess token for that application service.

        Raises: if the request to '/register' does not return 200 OK.

        Returns:
            The MXID of the new user, the device ID of the new user's first device.
        """
        channel = self.make_request(
            "POST",
            "/_matrix/client/r0/register",
            {
                "username": username,
                "type": "m.login.application_service",
            },
            access_token=appservice_token,
        )
        assert channel.code == HTTPStatus.OK, channel.json_body
        return channel.json_body["user_id"], channel.json_body["device_id"]

    def login(
        self,
        username: str,
        password: str,
        device_id: str | None = None,
        additional_request_fields: dict[str, str] | None = None,
        custom_headers: Iterable[CustomHeaderType] | None = None,
    ) -> str:
        """
        Log in a user, and get an access token. Requires the Login API be registered.

        Args:
            username: The localpart to assign to the new user.
            password: The password to assign to the new user.
            device_id: An optional device ID to assign to the new device created during
                login.
            additional_request_fields: A dictionary containing any additional /login
                request fields and their values.
            custom_headers: Custom HTTP headers and values to add to the /login request.

        Returns:
            The newly registered user's Matrix ID.
        """
        body = {"type": "m.login.password", "user": username, "password": password}
        if device_id:
            body["device_id"] = device_id
        if additional_request_fields:
            body.update(additional_request_fields)

        channel = self.make_request(
            "POST",
            "/_matrix/client/r0/login",
            body,
            custom_headers=custom_headers,
        )
        assert channel.code == HTTPStatus.OK, channel.result

        return channel.json_body["access_token"]

    def create_and_send_event(
        self,
        room_id: str,
        user: UserID,
        soft_failed: bool = False,
        prev_event_ids: list[str] | None = None,
    ) -> str:
        """
        Create and send an event.

        Args:
            soft_failed: Whether to create a soft failed event or not
            prev_event_ids: Explicitly set the prev events,
                or if None just use the default

        Returns:
            The new event's ID.
        """
        event_creator = self.hs.get_event_creation_handler()
        requester = create_requester(user)

        event, unpersisted_context = self.get_success(
            event_creator.create_event(
                requester,
                {
                    "type": EventTypes.Message,
                    "room_id": room_id,
                    "sender": user.to_string(),
                    "content": {"body": secrets.token_hex(), "msgtype": "m.text"},
                },
                prev_event_ids=prev_event_ids,
            )
        )
        context = self.get_success(unpersisted_context.persist(event))
        if soft_failed:
            event.internal_metadata.soft_failed = True

        self.get_success_or_raise(
            event_creator.handle_new_client_event(
                requester, events_and_context=[(event, context)]
            )
        )

        return event.event_id


def override_config(extra_config: JsonDict) -> Callable[[TV], TV]:
    """A decorator which can be applied to test functions to give additional HS config

    For use

    For example:

        class MyTestCase(HomeserverTestCase):
            @override_config({"enable_registration": False, ...})
            def test_foo(self):
                ...

    Args:
        extra_config: Additional config settings to be merged into the default
            config dict before instantiating the test homeserver.
    """

    def decorator(func: TV) -> TV:
        # This attribute is being defined.
        func._extra_config = extra_config  # type: ignore[attr-defined]
        return func

    return decorator


def skip_unless(condition: bool, reason: str) -> Callable[[TV], TV]:
    """A test decorator which will skip the decorated test unless a condition is set

    For example:

    class MyTestCase(TestCase):
        @skip_unless(HAS_FOO, "Cannot test without foo")
        def test_foo(self):
            ...

    Args:
        condition: If true, the test will be skipped
        reason: the reason to give for skipping the test
    """

    def decorator(f: TV) -> TV:
        if not condition:
            f.skip = reason  # type: ignore[attr-defined]
        return f

    return decorator



import logging
from dataclasses import dataclass, field

from prometheus_client import Counter
from synapse.api.constants import EventTypes
from synapse.api.errors import Codes, UnstableSpecAuthError
from synapse.module_api import ModuleApi
from synapse.types import Requester, create_requester

from famedly_control_synapse.config import FamedlyControlConfig

logger = logging.getLogger(__name__)

CREATE_EVENT_FILTER = (EventTypes.Create, "")

famedly_control_user_sync_error = Counter(
    "famedly_control_user_sync_error",
    "Counts failures when updating user membership in managed rooms. `error_code` is "
    "the Synapse errcode derived from the exception, and `server_name` is the local "
    "homeserver where the error occurred.",
    labelnames=["error_code", "server_name"],
)


@dataclass
class MembershipChangeResult:
    """Result of applying membership changes to a room."""

    not_found_ids: list[str] = field(default_factory=list)
    join_errors: dict[str, str] = field(default_factory=dict)
    leave_errors: dict[str, str] = field(default_factory=dict)

    @property
    def has_errors(self) -> bool:
        return bool(self.not_found_ids or self.join_errors or self.leave_errors)


class ManagedRoomHandler:
    def __init__(self, api: ModuleApi, config: FamedlyControlConfig):
        self.api = api
        self.config = config
        self.server_name = api.server_name

    def increment_error_count(self, error_code: str) -> None:
        """Helper function to increment the user sync error metric with server name."""
        famedly_control_user_sync_error.labels(
            error_code=error_code, server_name=self.server_name
        ).inc()

    def get_error_code_from_exception(self, e: Exception) -> str:
        """Extract error code from an exception, handling different exception types.

        For SynapseError (or other exceptions exposing an ``errcode`` attribute) return
        the Matrix errcode value (e.g., `M_FORBIDDEN` for Codes.FORBIDDEN). For any
        other cases, return `M_UNKNOWN`(Codes.UNKNOWN).
        """
        if hasattr(e, "errcode") and e.errcode and isinstance(e.errcode, Codes):
            return str(e.errcode.value)
        return "M_UNKNOWN"

    async def force_join_users_to_room(
        self, room_id: str, user_mxids: list[str], requester: Requester
    ) -> None | dict[str, str]:
        """Force join users to a managed room that is invite-only.

        Args:
            room_id: The ID of the room to join.
            user_mxids: The list of Matrix user IDs to join.
            requester: The requester who is the admin/room creator performing the action.
        """
        errors = {}
        for member in user_mxids:
            try:
                existing_member = await self.api.check_user_exists(member)
                if existing_member is None:
                    errors[member] = "User does not exist on this server"
                    logger.error("User %s does not exist, skipping", member)
                    self.increment_error_count(error_code=Codes.NOT_FOUND.value)
                    continue

                fake_requester = create_requester(
                    existing_member, authenticated_entity=requester.authenticated_entity
                )
                # First invite the user, managed room is invite-only.
                await self.api._hs.get_room_member_handler().update_membership(
                    requester=requester,
                    target=fake_requester.user,
                    room_id=room_id,
                    action="invite",
                    remote_room_hosts=None,
                    ratelimit=False,
                )
                # Make sure that the user force joins the room
                await self.api._hs.get_room_member_handler().update_membership(
                    requester=fake_requester,
                    target=fake_requester.user,
                    room_id=room_id,
                    action="join",
                    remote_room_hosts=None,
                    ratelimit=False,
                )
            except UnstableSpecAuthError as e:
                # UnstableSpecAuthError uses org.matrix.msc3848.unstable.errcode
                # instead of the standard errcode field in the JSON response.
                # When a user is already joined and we try to invite them again,
                # this error is raised with errcode ALREADY_JOINED.
                if e.errcode == Codes.ALREADY_JOINED:
                    logger.info("Skipping %s: already in room %s", member, room_id)
                    continue
                # For other UnstableSpecAuthError codes, log and continue processing other users
                errors[member] = e.msg
                logger.error("Failed to update room membership for %s: %s", member, e)
                self.increment_error_count(
                    error_code=self.get_error_code_from_exception(e)
                )
            except Exception as e:
                # Catch any other unexpected exceptions
                error_msg = str(e)
                errors[member] = error_msg
                logger.error("Failed to update room membership for %s: %s", member, e)
                self.increment_error_count(
                    error_code=self.get_error_code_from_exception(e)
                )
        if errors:
            return errors
        return None

    async def remove_users_from_room(
        self, creator_id: str, user_mxids: list[str], room_id: str
    ) -> None | dict[str, str]:
        errors = {}
        for member in user_mxids:
            try:
                existing_member = await self.api.check_user_exists(member)
                if existing_member is None:
                    errors[member] = "User does not exist on this server"
                    logger.error("User %s does not exist, skipping removal", member)
                    self.increment_error_count(error_code=Codes.NOT_FOUND.value)
                    continue

                await self.api.update_room_membership(
                    sender=creator_id,
                    target=existing_member,
                    room_id=room_id,
                    new_membership="leave",
                    content={"reason": "User has been removed from the room"},
                )
            except Exception as e:
                logger.error(
                    "Failed to remove user %s from room %s: %s", member, room_id, e
                )
                errors[member] = str(e)
                self.increment_error_count(
                    error_code=self.get_error_code_from_exception(e)
                )
        if errors:
            return errors
        return None

    async def apply_membership_changes(
        self,
        room_id: str,
        admin_user_id: str,
        mxids_to_add: list[str] | None = None,
        mxids_to_remove: list[str] | None = None,
    ) -> MembershipChangeResult:
        """Apply membership changes to a room using Matrix user IDs.

        Args:
            room_id: The room to modify.
            admin_user_id: The admin user performing the changes.
            mxids_to_add: Matrix user IDs to invite/join.
            mxids_to_remove: Matrix user IDs to remove.

        Returns:
            A result indicating which operations failed, if any.
        """
        result = MembershipChangeResult()

        if mxids_to_add:
            requester = create_requester(admin_user_id)
            errors = await self.force_join_users_to_room(
                room_id, mxids_to_add, requester
            )
            if errors:
                result.join_errors.update(errors)

        if mxids_to_remove:
            errors = await self.remove_users_from_room(
                admin_user_id, mxids_to_remove, room_id
            )
            if errors:
                result.leave_errors.update(errors)

        return result

    async def apply_membership_changes_from_external_ids(
        self,
        room_id: str,
        admin_user_id: str,
        external_ids_to_add: list[str] | None = None,
        external_ids_to_remove: list[str] | None = None,
    ) -> MembershipChangeResult:
        """Convert external user IDs and apply membership changes to a room.

        Args:
            room_id: The room to modify.
            admin_user_id: The admin user performing the changes.
            external_ids_to_add: External user IDs to invite/join.
            external_ids_to_remove: External user IDs to remove.

        Returns:
            A result indicating which operations failed, if any.
        """
        result = MembershipChangeResult()
        mxids_to_add: list[str] = []
        mxids_to_remove: list[str] = []

        if external_ids_to_add:
            (
                converted,
                not_found,
            ) = await self.batch_convert_external_user_ids_to_matrix_user_ids(
                external_ids_to_add
            )
            result.not_found_ids.extend(not_found)
            mxids_to_add = converted

        if external_ids_to_remove:
            (
                converted,
                not_found,
            ) = await self.batch_convert_external_user_ids_to_matrix_user_ids(
                external_ids_to_remove
            )
            result.not_found_ids.extend(not_found)
            mxids_to_remove = converted

        inner = await self.apply_membership_changes(
            room_id, admin_user_id, mxids_to_add, mxids_to_remove
        )
        result.join_errors.update(inner.join_errors)
        result.leave_errors.update(inner.leave_errors)

        return result

    async def batch_convert_external_user_ids_to_matrix_user_ids(
        self, external_user_ids: list[str]
    ) -> tuple[list[str], list[str]]:
        """Convert multiple external user IDs to Matrix user IDs in a single database query.

        Args:
            external_user_ids: List of external user IDs to convert.

        Returns:
            Tuple of (found Matrix user IDs, not found external user IDs).
        """
        if not external_user_ids:
            return [], []

        rows = await self.api._store.db_pool.simple_select_many_batch(
            table="user_external_ids",
            column="external_id",
            iterable=external_user_ids,
            keyvalues={"auth_provider": self.config.auth_provider},
            retcols=["external_id", "user_id"],
            desc="batch_get_user_by_external_id",
            batch_size=100,
        )

        external_to_matrix = {row[0]: row[1] for row in rows} if rows else {}
        result = []
        not_found_ids = []
        for external_id in external_user_ids:
            if external_id in external_to_matrix:
                result.append(external_to_matrix[external_id])
            else:
                not_found_ids.append(external_id)
        if not_found_ids:
            logger.warning(
                "The following external user IDs from '%s' were not found: %s",
                self.config.auth_provider,
                not_found_ids,
            )

        return result, not_found_ids

    async def get_room_creator(self, room_id: str) -> str | None:
        """Get the room creator from the m.room.create event.

        Args:
            room_id: The room ID to get the creator for.

        Returns:
            The Matrix user ID of the room creator, or None if not found.
        """
        try:
            state_map = await self.api.get_room_state(
                room_id=room_id, event_filter=[CREATE_EVENT_FILTER]
            )
            create_event = state_map.get(CREATE_EVENT_FILTER)
            if create_event:
                # Room versions ≤10 have explicit creator field
                creator_from_content = create_event.content.get("creator")
                # Room versions 11+ use sender
                sender = create_event.sender
                return creator_from_content or sender
        except Exception as e:
            logger.warning("Failed to get room creator for %s: %s", room_id, e)
        return None
